#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2006 Jean-Francois Barraud, barraud@math.univ-lille1.fr
#               2021 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# barraud@math.univ-lille1.fr
#
"""
This code defines a basic class (PathModifier) of effects whose purpose is
to somehow deform given objects: one common tasks for all such effect is to
convert shapes, groups, clones to paths. The class has several functions to
make this (more or less!) easy.
As an example, a second class (Diffeo) is derived from it,
to implement deformations of the form X=f(x,y), Y=g(x,y)...
"""

import inkex
from inkex import PathElement, Group, Use
from inkex.bezier import pointdistance, beziersplitatt

# This deprecated API is used by some external extensions.
from inkex.deprecated import zSort  # pylint: disable=unused-import


class PathModifier(inkex.EffectExtension):
    """Select list manipulation"""

    def expand_groups(self, elements, transferTransform=True):
        for node_id, node in list(elements.items()):
            if isinstance(node, inkex.Group):
                mat = node.transform
                for child in node:
                    if transferTransform:
                        child.transform = mat @ child.transform
                    elements.update(self.expand_groups({child.get("id"): child}))
                if transferTransform and node.get("transform"):
                    del node.attrib["transform"]
                # Group is now replaced, so remove it.
                elements.pop(node_id)
        return elements

    def expand_clones(self, elements, transferTransform=True, replace=True):
        for node_id, node in list(elements.items()):
            if isinstance(node, Group):
                self.expand_groups(elements, transferTransform)
                self.expand_clones(elements, transferTransform, replace)
                # Hum... not very efficient if there are many clones of groups...

            elif isinstance(node, Use):
                newnode = node.unlink()
                elements.pop(node_id)
                newid = newnode.get("id")
                elements.update(
                    self.expand_clones({newid: newnode}, transferTransform, replace)
                )
        return elements

    def objects_to_paths(self, elements, replace=True):
        """Replace all non-paths with path objects"""
        for node in list(elements.values()):
            elem = node.to_path_element()
            if replace:
                node.replace_with(elem)
                elem.set("id", node.get("id"))
            elements[elem.get("id")] = elem

    def effect(self):
        raise NotImplementedError("overwrite this method in subclasses")
        self.objects_to_paths(self.svg.selection, True)
        self.bbox = self.svg.selection.bounding_box()
        for node in self.svg.selection.filter(PathElement):
            path = node.path.to_superpath()
            # do what ever you want with "path"!
            node.path = path

    @staticmethod
    def lengthtotime(l, lengths, isclosed):
        """
        Receives an arc length l, and returns the index of the segment in skelcomp
        containing the corresponding point, to gether with the position of the point on this segment.

        If the deformer is closed, do computations modulo the toal length.
        """
        if isclosed:
            l = l % sum(lengths)
        if l <= 0:
            return 0, l / lengths[0]
        i = 0
        while (i < len(lengths)) and (lengths[i] <= l):
            l -= lengths[i]
            i += 1
        t = l / lengths[min(i, len(lengths) - 1)]
        return i, t

    @staticmethod
    def flipxy(path):
        """Swaps x and y coordinate of all path vertices"""
        for pathcomp in path:
            for ctl in pathcomp:
                for pt in ctl:
                    tmp = pt[0]
                    pt[0] = -pt[1]
                    pt[1] = -tmp

    @staticmethod
    def offset(pathcomp, dx, dy):
        """Shifts a subpath by (dx, dy)"""
        for ctl in pathcomp:
            for pt in ctl:
                pt[0] += dx
                pt[1] += dy

    @staticmethod
    def stretch(pathcomp, xscale, yscale, org):
        """Stretches a subpath by (xscale, yscale) relative to origin org"""
        for ctl in pathcomp:
            for pt in ctl:
                pt[0] = org[0] + (pt[0] - org[0]) * xscale
                pt[1] = org[1] + (pt[1] - org[1]) * yscale

    @staticmethod
    def linearize(p, tolerance=0.001):
        """
        This function receives a component of a 'cubicsuperpath' and returns two things:
        The path subdivided in many straight segments, and an array containing the length of each segment.

        We could work with bezier path as well, but bezier arc lengths are (re)computed for each point
        in the deformed object. For complex paths, this might take a while.
        """
        zero = 0.000001
        i = 0
        d = 0
        lengths = []
        while i < len(p) - 1:
            box = pointdistance(p[i][1], p[i][2])
            box += pointdistance(p[i][2], p[i + 1][0])
            box += pointdistance(p[i + 1][0], p[i + 1][1])
            chord = pointdistance(p[i][1], p[i + 1][1])
            if (box - chord) > tolerance:
                b1, b2 = beziersplitatt(
                    [p[i][1], p[i][2], p[i + 1][0], p[i + 1][1]], 0.5
                )
                p[i][2][0], p[i][2][1] = b1[1]
                p[i + 1][0][0], p[i + 1][0][1] = b2[2]
                p.insert(
                    i + 1,
                    [[b1[2][0], b1[2][1]], [b1[3][0], b1[3][1]], [b2[1][0], b2[1][1]]],
                )
            else:
                d = (box + chord) / 2
                lengths.append(d)
                i += 1
        new = [p[i][1] for i in range(0, len(p) - 1) if lengths[i] > zero]
        new.append(p[-1][1])
        lengths = [l for l in lengths if l > zero]
        return new, lengths

    def get_patterns_and_skeletons(self, expand_patterns=False, duplicate=False):
        """Gets the pattern node and skeleton (with applied transforms) from selection"""
        # first selected->pattern, all but first selected-> skeletons
        skeletons = self.svg.selection.rendering_order()

        elem = skeletons.pop()
        if duplicate:
            elem = elem.duplicate()

        if expand_patterns:
            patterns = {elem.get_id(): elem}
            self.expand_clones(patterns, True, False)
            self.objects_to_paths(patterns)
        else:
            patterns = elem

        self.expand_clones(skeletons, True, False)
        self.objects_to_paths(skeletons)
        return patterns, skeletons.id_dict()


class Diffeo(PathModifier):
    def applyDiffeo(self, bpt, vects=()):
        # bpt is a base point and for v in vectors, v'=v-p is a tangent vector at bpt.
        # Defaults to identity!
        for v in vects:
            v[0] -= bpt[0]
            v[1] -= bpt[1]

        # -- your transformations go here:
        # x,y=bpt
        # bpt[0]=f(x,y)
        # bpt[1]=g(x,y)
        # for v in vects:
        #    vx,vy=v
        #    v[0]=df/dx(x,y)*vx+df/dy(x,y)*vy
        #    v[1]=dg/dx(x,y)*vx+dg/dy(x,y)*vy
        #
        # -- !caution! y-axis is pointing downward!

        for v in vects:
            v[0] += bpt[0]
            v[1] += bpt[1]

    def effect(self):
        self.expand_clones(self.svg.selection, True)
        self.expand_groups(self.svg.selection, True)
        self.objects_to_paths(self.svg.selection, True)
        self.bbox = self.svg.selection.bounding_box()
        for node in self.svg.selection.filter(PathElement):
            path = node.path.to_superpath()
            for sub in path:
                for ctlpt in sub:
                    self.applyDiffeo(ctlpt[1], (ctlpt[0], ctlpt[2]))
            node.path = path



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2014  Ryan Lerch     (multiple difference)
#              2016  Maren Hachmann <marenhachmannATyahoo.com> (refactoring, extend to multibool)
#              2017  Alvin Penner   <penner@vaxxine.com> (apply to 'File Save As...')
#
# This code is based on 'inkscape-extension-multiple-difference' by Ryan Lerch
# see : https://github.com/ryanlerch/inkscape-extension-multiple-difference
# also: https://github.com/Moini/inkscape-extensions-multi-bool
# It will call up a new instance of Inkscape and process the image there,
# so that the original file is left intact.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This extension will pre-process a vector image by applying the operations:
'EditSelectAllInAllLayers' and 'ObjectToPath'
before calling the dialog File->Save As....
"""

import inkex
from inkex.base import TempDirMixin
from inkex.command import inkscape_command
from inkex import load_svg


class PreProcess(TempDirMixin, inkex.EffectExtension):
    def effect(self):
        self.document = load_svg(
            inkscape_command(
                self.svg,
                verbs=["EditSelectAllInAllLayers", "EditUnlinkClone", "ObjectToPath"],
            )
        )


if __name__ == "__main__":
    PreProcess().run()



# coding=utf-8
#
# Copyright (C) 2006 Jos Hirth, kaioa.com
# Copyright (C) 2007 Aaron C. Spike
# Copyright (C) 2009 Monash University
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Basic color controls
"""


# All the names that get added to the inkex API itself.
__all__ = ("Color", "ColorError", "ColorIdError")

SVG_COLOR = {
    "aliceblue": "#f0f8ff",
    "antiquewhite": "#faebd7",
    "aqua": "#00ffff",
    "aquamarine": "#7fffd4",
    "azure": "#f0ffff",
    "beige": "#f5f5dc",
    "bisque": "#ffe4c4",
    "black": "#000000",
    "blanchedalmond": "#ffebcd",
    "blue": "#0000ff",
    "blueviolet": "#8a2be2",
    "brown": "#a52a2a",
    "burlywood": "#deb887",
    "cadetblue": "#5f9ea0",
    "chartreuse": "#7fff00",
    "chocolate": "#d2691e",
    "coral": "#ff7f50",
    "cornflowerblue": "#6495ed",
    "cornsilk": "#fff8dc",
    "crimson": "#dc143c",
    "cyan": "#00ffff",
    "darkblue": "#00008b",
    "darkcyan": "#008b8b",
    "darkgoldenrod": "#b8860b",
    "darkgray": "#a9a9a9",
    "darkgreen": "#006400",
    "darkgrey": "#a9a9a9",
    "darkkhaki": "#bdb76b",
    "darkmagenta": "#8b008b",
    "darkolivegreen": "#556b2f",
    "darkorange": "#ff8c00",
    "darkorchid": "#9932cc",
    "darkred": "#8b0000",
    "darksalmon": "#e9967a",
    "darkseagreen": "#8fbc8f",
    "darkslateblue": "#483d8b",
    "darkslategray": "#2f4f4f",
    "darkslategrey": "#2f4f4f",
    "darkturquoise": "#00ced1",
    "darkviolet": "#9400d3",
    "deeppink": "#ff1493",
    "deepskyblue": "#00bfff",
    "dimgray": "#696969",
    "dimgrey": "#696969",
    "dodgerblue": "#1e90ff",
    "firebrick": "#b22222",
    "floralwhite": "#fffaf0",
    "forestgreen": "#228b22",
    "fuchsia": "#ff00ff",
    "gainsboro": "#dcdcdc",
    "ghostwhite": "#f8f8ff",
    "gold": "#ffd700",
    "goldenrod": "#daa520",
    "gray": "#808080",
    "grey": "#808080",
    "green": "#008000",
    "greenyellow": "#adff2f",
    "honeydew": "#f0fff0",
    "hotpink": "#ff69b4",
    "indianred": "#cd5c5c",
    "indigo": "#4b0082",
    "ivory": "#fffff0",
    "khaki": "#f0e68c",
    "lavender": "#e6e6fa",
    "lavenderblush": "#fff0f5",
    "lawngreen": "#7cfc00",
    "lemonchiffon": "#fffacd",
    "lightblue": "#add8e6",
    "lightcoral": "#f08080",
    "lightcyan": "#e0ffff",
    "lightgoldenrodyellow": "#fafad2",
    "lightgray": "#d3d3d3",
    "lightgreen": "#90ee90",
    "lightgrey": "#d3d3d3",
    "lightpink": "#ffb6c1",
    "lightsalmon": "#ffa07a",
    "lightseagreen": "#20b2aa",
    "lightskyblue": "#87cefa",
    "lightslategray": "#778899",
    "lightslategrey": "#778899",
    "lightsteelblue": "#b0c4de",
    "lightyellow": "#ffffe0",
    "lime": "#00ff00",
    "limegreen": "#32cd32",
    "linen": "#faf0e6",
    "magenta": "#ff00ff",
    "maroon": "#800000",
    "mediumaquamarine": "#66cdaa",
    "mediumblue": "#0000cd",
    "mediumorchid": "#ba55d3",
    "mediumpurple": "#9370db",
    "mediumseagreen": "#3cb371",
    "mediumslateblue": "#7b68ee",
    "mediumspringgreen": "#00fa9a",
    "mediumturquoise": "#48d1cc",
    "mediumvioletred": "#c71585",
    "midnightblue": "#191970",
    "mintcream": "#f5fffa",
    "mistyrose": "#ffe4e1",
    "moccasin": "#ffe4b5",
    "navajowhite": "#ffdead",
    "navy": "#000080",
    "oldlace": "#fdf5e6",
    "olive": "#808000",
    "olivedrab": "#6b8e23",
    "orange": "#ffa500",
    "orangered": "#ff4500",
    "orchid": "#da70d6",
    "palegoldenrod": "#eee8aa",
    "palegreen": "#98fb98",
    "paleturquoise": "#afeeee",
    "palevioletred": "#db7093",
    "papayawhip": "#ffefd5",
    "peachpuff": "#ffdab9",
    "peru": "#cd853f",
    "pink": "#ffc0cb",
    "plum": "#dda0dd",
    "powderblue": "#b0e0e6",
    "purple": "#800080",
    "rebeccapurple": "#663399",
    "red": "#ff0000",
    "rosybrown": "#bc8f8f",
    "royalblue": "#4169e1",
    "saddlebrown": "#8b4513",
    "salmon": "#fa8072",
    "sandybrown": "#f4a460",
    "seagreen": "#2e8b57",
    "seashell": "#fff5ee",
    "sienna": "#a0522d",
    "silver": "#c0c0c0",
    "skyblue": "#87ceeb",
    "slateblue": "#6a5acd",
    "slategray": "#708090",
    "slategrey": "#708090",
    "snow": "#fffafa",
    "springgreen": "#00ff7f",
    "steelblue": "#4682b4",
    "tan": "#d2b48c",
    "teal": "#008080",
    "thistle": "#d8bfd8",
    "tomato": "#ff6347",
    "turquoise": "#40e0d0",
    "violet": "#ee82ee",
    "wheat": "#f5deb3",
    "white": "#ffffff",
    "whitesmoke": "#f5f5f5",
    "yellow": "#ffff00",
    "yellowgreen": "#9acd32",
    "none": None,
}
COLOR_SVG = {value: name for name, value in SVG_COLOR.items()}


def is_color(color):
    """Determine if it is a color that we can use. If not, leave it unchanged."""
    try:
        return bool(Color(color))
    except ColorError:
        return False


def constrain(minim, value, maxim, channel):
    """Returns the value so long as it is between min and max values"""
    if channel == "h":  # Hue
        return value % maxim  # Wrap around hue value
    return min([maxim, max([minim, value])])


class ColorError(KeyError):
    """Specific color parsing error"""


class ColorIdError(ColorError):
    """Special color error for gradient and color stop ids"""


class Color(list):
    """An RGB array for the color

    Can be constructed from valid CSS color attributes, as well as
    tuple/list + color space. Percentage values are supported.

    .. versionchanged:: 1.2
        Clarification with respect to values denoting unity: For RGB color channels,
        "1.0", 1.0 and "100%" are treated as 255, while "1" and 1 are treated as 1.

    """

    red = property(
        lambda self: self.to_rgb()[0], lambda self, value: self._set(0, value)
    )
    green = property(
        lambda self: self.to_rgb()[1], lambda self, value: self._set(1, value)
    )
    blue = property(
        lambda self: self.to_rgb()[2], lambda self, value: self._set(2, value)
    )
    alpha = property(
        lambda self: self.to_rgba()[3],
        lambda self, value: self._set(3, value, ("rgba",)),
    )
    hue = property(
        lambda self: self.to_hsl()[0], lambda self, value: self._set(0, value, ("hsl",))
    )
    saturation = property(
        lambda self: self.to_hsl()[1], lambda self, value: self._set(1, value, ("hsl",))
    )
    lightness = property(
        lambda self: self.to_hsl()[2], lambda self, value: self._set(2, value, ("hsl",))
    )

    def __init__(self, color=None, space="rgb"):
        super().__init__()
        if isinstance(color, Color):
            space, color = color.space, list(color)

        if isinstance(color, str):
            # String from xml or css attributes
            space, color = self.parse_str(color.strip())

        if isinstance(color, int):
            # Number from arg parser colour value
            space, color = self.parse_int(color)

        # Empty list means 'none', or no color
        if color is None:
            color = []

        if not isinstance(color, (list, tuple)):
            raise ColorError("Not a known a color value")

        self.space = space
        try:
            for val in color:
                self.append(val)
        except ValueError as error:
            raise ColorError("Bad color list") from error

    def __hash__(self):
        """Allow colors to be hashable"""
        return tuple(self.to_rgba()).__hash__()

    def _set(self, index, value, spaces=("rgb", "rgba")):
        """Set the color value in place, limits setter to specific color space"""
        # Named colors are just rgb, so dump name memory
        if self.space == "named":
            self.space = "rgb"
        if not self.space in spaces:
            if index == 3 and self.space == "rgb":
                # Special, add alpha, don't convert back to rgb
                self.space = "rgba"
                self.append(constrain(0.0, float(value), 1.0, "a"))
                return
            # Set in other colour space and convert back and forth
            target = self.to(spaces[0])
            target[index] = constrain(0, int(value), 255, spaces[0][index])
            self[:] = target.to(self.space)
            return
        self[index] = constrain(0, int(value), 255, spaces[0][index])

    def append(self, val):
        """Append a value to the local list"""
        if len(self) == len(self.space):
            raise ValueError("Can't add any more values to color.")

        if isinstance(val, str):
            val = val.strip()
            if val.endswith("%"):
                val = float(val.strip("%")) / 100
            elif "." in val:
                val = float(val)
            else:
                val = int(val)

        end_type = int
        if len(self) == 3:  # Alpha value
            val = min([1.0, val])
            end_type = float
        elif isinstance(val, float) and val <= 1.0:
            val *= 255

        if isinstance(val, (int, float)):
            super().append(max(end_type(val), 0))

    @staticmethod
    def parse_str(color):
        """Creates a rgb int array"""
        # Handle pre-defined svg color values
        if color and color.lower() in SVG_COLOR:
            return "named", Color.parse_str(SVG_COLOR[color.lower()])[1]

        if color is None:
            return "rgb", None

        if color.startswith("url("):
            raise ColorIdError("Color references other element id, e.g. a gradient")

        # Next handle short colors (css: #abc -> #aabbcc)
        if color.startswith("#"):
            # Remove any icc or ilab directives
            # FUTURE: We could use icc or ilab information
            col = color.split(" ")[0]
            if len(col) == 4:
                # pylint: disable=consider-using-f-string
                col = "#{1}{1}{2}{2}{3}{3}".format(*col)

            # Convert hex to integers
            try:
                return "rgb", (int(col[1:3], 16), int(col[3:5], 16), int(col[5:], 16))
            except ValueError as error:
                raise ColorError(f"Bad RGB hex color value {col}") from error

        # Handle other css color values
        elif "(" in color and ")" in color:
            space, values = color.lower().strip().strip(")").split("(")
            return space, values.split(",")

        try:
            return Color.parse_int(int(color))
        except ValueError:
            pass

        raise ColorError(f"Unknown color format: {color}")

    @staticmethod
    def parse_int(color):
        """Creates an rgb or rgba from a long int"""
        space = "rgb"
        color = [
            ((color >> 24) & 255),  # red
            ((color >> 16) & 255),  # green
            ((color >> 8) & 255),  # blue
            ((color & 255) / 255.0),  # opacity
        ]
        if color[-1] == 1.0:
            color.pop()
        else:
            space = "rgba"
        return space, color

    def __str__(self):
        """int array to #rrggbb"""
        # pylint: disable=consider-using-f-string
        if not self:
            return "none"
        if self.space == "named":
            rgbhex = "#{0:02x}{1:02x}{2:02x}".format(*self)
            if rgbhex in COLOR_SVG:
                return COLOR_SVG[rgbhex]
            self.space = "rgb"
        if self.space == "rgb":
            return "#{0:02x}{1:02x}{2:02x}".format(*self)
        if self.space == "rgba":
            if self[3] == 1.0:
                return "rgb({:g}, {:g}, {:g})".format(*self[:3])
            return "rgba({:g}, {:g}, {:g}, {:g})".format(*self)
        if self.space == "hsl":
            return "hsl({0:g}, {1:g}, {2:g})".format(*self)
        raise ColorError(f"Can't print colour space '{self.space}'")

    def __int__(self):
        """int array to large integer"""
        if not self:
            return -1
        color = self.to_rgba()
        return (
            (color[0] << 24)
            + (color[1] << 16)
            + (color[2] << 8)
            + (int(color[3] * 255))
        )

    def to(self, space):  # pylint: disable=invalid-name
        """Dynamic caller for to_hsl, to_rgb, etc"""
        return getattr(self, "to_" + space)()

    def to_hsl(self):
        """Turn this color into a Hue/Saturation/Lightness colour space"""
        if not self and self.space in ("rgb", "named"):
            return self.to_rgb().to_hsl()
        if self.space == "hsl":
            return self
        if self.space in ("named"):
            return self.to_rgb().to_hsl()
        if self.space == "rgb":
            return Color(rgb_to_hsl(*self.to_floats()), space="hsl")
        raise ColorError(f"Unknown color conversion {self.space}->hsl")

    def to_rgb(self):
        """Turn this color into a Red/Green/Blue colour space"""
        if not self and self.space in ("rgb", "named"):
            return Color([0, 0, 0])
        if self.space == "rgb":
            return self
        if self.space in ("rgba", "named"):
            return Color(self[:3], space="rgb")
        if self.space == "hsl":
            return Color(hsl_to_rgb(*self.to_floats()), space="rgb")
        raise ColorError(f"Unknown color conversion {self.space}->rgb")

    def to_rgba(self, alpha=1.0):
        """Turn this color isn't an RGB with Alpha colour space"""
        if self.space == "rgba":
            return self
        return Color(self.to_rgb() + [alpha], "rgba")

    def to_floats(self):
        """Returns the colour values as percentage floats (0.0 - 1.0)"""
        return [val / 255.0 for val in self]

    def to_named(self):
        """Convert this color to a named color if possible"""
        if not self:
            return Color()
        return Color(COLOR_SVG.get(str(self), str(self)))

    def interpolate(self, other, fraction):
        """Interpolate two colours by the given fraction

        .. versionadded:: 1.1"""
        from .tween import ColorInterpolator  # pylint: disable=import-outside-toplevel

        return ColorInterpolator(self, other).interpolate(fraction)

    @staticmethod
    def isnone(x):
        """Checks if a given color is none

        .. versionadded:: 1.2"""

        if x is None or (isinstance(x, str) and x.lower() == "none"):
            return True
        return False

    @staticmethod
    def iscolor(x, accept_none=False):
        """Checks if a given value can be parsed as a color

        .. versionadded:: 1.2"""
        if isinstance(x, str) and (accept_none or not (Color.isnone(x))):
            try:
                Color(x)
                return True
            except (ColorError):
                pass
        if isinstance(x, Color):
            return True
        return False


def rgb_to_hsl(red, green, blue):
    """RGB to HSL colour conversion"""
    rgb_max = max(red, green, blue)
    rgb_min = min(red, green, blue)
    delta = rgb_max - rgb_min
    hsl = [0.0, 0.0, (rgb_max + rgb_min) / 2.0]
    if delta != 0:
        if hsl[2] <= 0.5:
            hsl[1] = delta / (rgb_max + rgb_min)
        else:
            hsl[1] = delta / (2 - rgb_max - rgb_min)

        if red == rgb_max:
            hsl[0] = (green - blue) / delta
        elif green == rgb_max:
            hsl[0] = 2.0 + (blue - red) / delta
        elif blue == rgb_max:
            hsl[0] = 4.0 + (red - green) / delta

        hsl[0] /= 6.0
        if hsl[0] < 0:
            hsl[0] += 1
        if hsl[0] > 1:
            hsl[0] -= 1
    return hsl


def hsl_to_rgb(hue, sat, light):
    """HSL to RGB Color Conversion"""
    if sat == 0:
        return [light, light, light]  # Gray

    if light < 0.5:
        val2 = light * (1 + sat)
    else:
        val2 = light + sat - light * sat
    val1 = 2 * light - val2
    return [
        _hue_to_rgb(val1, val2, hue * 6 + 2.0),
        _hue_to_rgb(val1, val2, hue * 6),
        _hue_to_rgb(val1, val2, hue * 6 - 2.0),
    ]


def _hue_to_rgb(val1, val2, hue):
    if hue < 0:
        hue += 6.0
    if hue > 6:
        hue -= 6.0
    if hue < 1:
        return val1 + (val2 - val1) * hue
    if hue < 3:
        return val2
    if hue < 4:
        return val1 + (val2 - val1) * (4 - hue)
    return val1



#!/usr/bin/env python
# coding=utf-8
"""
Test elements extra logic from svg xml lxml custom classes.
"""

from ..utils import PY3
from ..inx import InxFile

INTERNAL_ARGS = ("help", "output", "id", "selected-nodes")
ARG_TYPES = {
    "Boolean": "bool",
    "Color": "color",
    "str": "string",
    "int": "int",
    "float": "float",
}


class InxMixin:
    """Tools for Testing INX files, use as a mixin class:

    class MyTests(InxMixin, TestCase):
        def test_inx_file(self):
            self.assertInxIsGood("some_inx_file.inx")
    """

    def assertInxIsGood(self, inx_file):  # pylint: disable=invalid-name
        """Test the inx file for consistancy and correctness"""
        self.assertTrue(PY3, "INX files can only be tested in python3")

        inx = InxFile(inx_file)
        if "help" in inx.ident or inx.script.get("interpreter", None) != "python":
            return
        cls = inx.extension_class
        # Check class can be matched in python file
        self.assertTrue(cls, f"Can not find class for {inx.filename}")
        # Check name is reasonable for the class
        if not cls.multi_inx:
            self.assertEqual(
                cls.__name__,
                inx.slug,
                f"Name of extension class {cls.__module__}.{cls.__name__} "
                f"is different from ident {inx.slug}",
            )
            self.assertParams(inx, cls)

    def assertParams(self, inx, cls):  # pylint: disable=invalid-name
        """Confirm the params in the inx match the python script

        .. versionchanged:: 1.2
            Also checks that the default values are identical"""
        params = {param.name: self.parse_param(param) for param in inx.params}
        args = dict(self.introspect_arg_parser(cls().arg_parser))
        mismatch_a = list(set(params) ^ set(args) & set(params))
        mismatch_b = list(set(args) ^ set(params) & set(args))
        self.assertFalse(
            mismatch_a, f"{inx.filename}: Inx params missing from arg parser"
        )
        self.assertFalse(
            mismatch_b, f"{inx.filename}: Script args missing from inx xml"
        )

        for param in args:
            if params[param]["type"] and args[param]["type"]:
                self.assertEqual(
                    params[param]["type"],
                    args[param]["type"],
                    f"Type is not the same for {inx.filename}:param:{param}",
                )
            inxdefault = params[param]["default"]
            argsdefault = args[param]["default"]
            if inxdefault and argsdefault:
                # for booleans, the inx is lowercase and the param is uppercase
                if params[param]["type"] == "bool":
                    argsdefault = str(argsdefault).lower()
                elif params[param]["type"] not in ["string", None, "color"] or args[
                    param
                ]["type"] in ["int", "float"]:
                    # try to parse the inx value to compare numbers to numbers
                    inxdefault = float(inxdefault)
                if args[param]["type"] == "color" or callable(args[param]["default"]):
                    # skip color, method types
                    continue
                self.assertEqual(
                    argsdefault,
                    inxdefault,
                    f"Default value is not the same for {inx.filename}:param:{param}",
                )

    def introspect_arg_parser(self, arg_parser):
        """Pull apart the arg parser to find out what we have in it"""
        for (
            action
        ) in arg_parser._optionals._actions:  # pylint: disable=protected-access
            for opt in action.option_strings:
                # Ignore params internal to inkscape (thus not in the inx)
                if opt.startswith("--") and opt[2:] not in INTERNAL_ARGS:
                    yield (opt[2:], self.introspect_action(action))

    @staticmethod
    def introspect_action(action):
        """Pull apart a single action to get at the juicy insides"""
        return {
            "type": ARG_TYPES.get((action.type or str).__name__, "string"),
            "default": action.default,
            "choices": action.choices,
            "help": action.help,
        }

    @staticmethod
    def parse_param(param):
        """Pull apart the param element in the inx file"""
        if param.param_type in ("optiongroup", "notebook"):
            options = param.options
            return {
                "type": None,
                "choices": options,
                "default": options and options[0] or None,
            }
        param_type = param.param_type
        if param.param_type in ("path",):
            param_type = "string"
        return {
            "type": param_type,
            "default": param.text,
            "choices": None,
        }



#!/usr/bin/env python
"""Increase satuation"""

import inkex


class MoreSaturation(inkex.ColorExtension):
    """Increase saturation of selected objects"""

    def modify_color(self, name, color):
        color.saturation += int(0.05 * 255.0)
        return color


if __name__ == "__main__":
    MoreSaturation().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2009 Aurelio A. Heckert, aurium (a) gmail dot com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Interpolation of attributes in selected objects or group's children.
"""

import inkex
from inkex.localization import inkex_gettext as _
from inkex.tween import ColorInterpolator, ValueInterpolator, AttributeInterpolator
from inkex.utils import is_number


class InterpAttG(inkex.EffectExtension):
    """
    This effect applies a value for any interpolatable attribute for all
    elements inside the selected group or for all elements in a multiple selection.
    """

    def __init__(self):
        super(InterpAttG, self).__init__()
        self.arg_parser.add_argument(
            "-a",
            "--att",
            type=str,
            dest="att",
            default="width",
            help="Attribute to be interpolated.",
        )
        self.arg_parser.add_argument(
            "-o",
            "--att-other",
            type=str,
            dest="att_other",
            help="Other attribute (for a limited UI).",
        )
        self.arg_parser.add_argument(
            "-t",
            "--att-other-type",
            type=self.arg_class([ColorInterpolator, ValueInterpolator]),
            dest="att_other_type",
            help="The other attribute type.",
        )
        self.arg_parser.add_argument(
            "-w",
            "--att-other-where",
            type=str,
            dest="att_other_where",
            default="tag",
            help="That is a tag attribute or a style attribute?",
        )
        self.arg_parser.add_argument(
            "-s",
            "--start-val",
            type=str,
            dest="start_val",
            default="#F00",
            help="Initial interpolation value.",
        )
        self.arg_parser.add_argument(
            "-e",
            "--end-val",
            type=str,
            dest="end_val",
            default="#00F",
            help="End interpolation value.",
        )
        self.arg_parser.add_argument(
            "-u", "--unit", type=str, dest="unit", default="none", help="Values unit."
        )
        self.arg_parser.add_argument(
            "--zsort",
            type=inkex.Boolean,
            dest="zsort",
            default=False,
            help="use z-order instead of selection order",
        )
        self.arg_parser.add_argument(
            "--tab",
            type=str,
            dest="tab",
            help="The selected UI-tab when OK was pressed",
        )

    def get_elements(self):
        """Returns a list of elements to work on"""
        if not self.svg.selection:
            return []

        if len(self.svg.selection) > 1:
            # multiple selection
            if self.options.zsort:
                return list(self.svg.selection.rendering_order().values())
            return list(self.svg.selection.values())

        # must be a group
        node = self.svg.selection.filter(inkex.Group).first()
        return list(node) or []

    def create_dummy_nodes(self, path):
        """Create dummy nodes to use the interpolation classes defined in inkex.tween"""
        pat1 = inkex.PathElement()
        pat2 = inkex.PathElement()

        start_value = self.options.start_val.replace(",", ".")
        end_value = self.options.end_val.replace(",", ".")
        if self.options.unit != "none":
            start_value += self.options.unit
            end_value += self.options.unit
        self.apply_value(pat1, path, start_value)
        self.apply_value(pat2, path, end_value)
        return pat1, pat2

    @staticmethod
    def apply_value(node, path, value):
        """Applies a value to a given node. If path starts with "transform/" or "style/", the
        value is applied to either transform or style."""
        if path.startswith("style/"):
            att_name = path[6:]
            node.style[att_name] = value
        elif path.startswith("transform/"):
            if not is_number(value):
                raise inkex.AbortExtension(
                    _("Unable to set attribute {} to {}").format(path, value)
                )
            if path == "transform/trans-x":
                node.transform.add_translate(value, 0)
            elif path == "transform/trans-y":
                node.transform.add_translate(0, value)
            elif path == "transform/scale":
                node.transform.add_scale(value)
        elif path == "transform":
            node.transform @= value
        else:
            node.set(path, value)

    def effect(self):
        method = None
        if self.options.att == "other":
            if self.options.att_other is None:
                raise inkex.AbortExtension(
                    _("You selected 'Other'. Please enter an attribute to interpolate.")
                )
            if self.options.att_other_where == "tag":
                path = self.options.att_other
            else:
                path = self.options.att_other_where + "/" + self.options.att_other
            method = self.options.att_other_type
        else:
            path = self.options.att
            if self.options.att == "height" or self.options.att == "width":
                method = inkex.tween.UnitValueInterpolator

        path1, path2 = self.create_dummy_nodes(path)
        if path.startswith("transform"):
            path = "transform"
        try:
            # maybe tween knows what do do with this attribute?
            interpolator = AttributeInterpolator.create_from_attribute(
                path1, path2, path, None
            )
        except:
            # okay, apparently not
            interpolator = AttributeInterpolator.create_from_attribute(
                path1, path2, path, method
            )
        collection = self.get_elements()
        if not collection:
            raise inkex.AbortExtension(_("There is no selection to interpolate"))

        steps = [1.0 / (len(collection) - 1) * i for i in range(len(collection))]

        for time, node in zip(steps, collection):
            new_value = interpolator.interpolate(time)
            InterpAttG.apply_value(node, path, new_value)

        return True


if __name__ == "__main__":
    InterpAttG().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2020 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Path To Absolute
"""
import inkex


class ToAbsolute(inkex.EffectExtension):
    """Convert any selected object to absolute/object-to-path/bezier only paths"""

    def effect(self):
        """Performs the effect."""
        for node in self.svg.selection.values():
            if not isinstance(node, inkex.PathElement):
                node = node.replace_with(node.to_path_element())
            node.path = node.path.to_absolute().to_superpath().to_path()


if __name__ == "__main__":
    ToAbsolute().run()



#!/usr/bin/env python
#
# Copyright 2008, 2009 Hannes Hochreiner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.
#
"""Uninstall jessyInk"""

import inkex
from jessyink_install import JessyInkMixin


class Uninstall(JessyInkMixin, inkex.EffectExtension):
    """Uninstall jessyInk from this svg"""

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--remove_script", type=inkex.Boolean, default=True)
        pars.add_argument("--remove_effects", type=inkex.Boolean, default=True)
        pars.add_argument("--remove_masterSlide", type=inkex.Boolean, default=True)
        pars.add_argument("--remove_transitions", type=inkex.Boolean, default=True)
        pars.add_argument("--remove_autoTexts", type=inkex.Boolean, default=True)
        pars.add_argument("--remove_views", type=inkex.Boolean, default=True)

    def effect(self):
        # Remove script, if so desired.
        if self.options.remove_script:
            # Find and delete script node.
            for node in self.svg.xpath("//svg:script[@id='JessyInk']"):
                node.delete()

            # Remove "jessyInkInit()" in the "onload" attribute, if present.
            prop_list = []
            if self.svg.get("onload"):
                prop_list = self.prop_str_to_list(self.svg.get("onload"))

            for prop in prop_list:
                if prop == "jessyInkInit()":
                    prop_list.remove("jessyInkInit()")

            if prop_list:
                self.svg.set("onload", self.list_to_prop_str(prop_list))
            else:
                if self.document.getroot().get("onload"):
                    del self.document.getroot().attrib["onload"]

        self.attr_remove("effectIn", self.options.remove_effects)
        self.attr_remove("effectOut", self.options.remove_effects)
        self.attr_remove("masterSlide", self.options.remove_masterSlide)
        self.attr_remove("transitionIn", self.options.remove_transitions)
        self.attr_remove("transitionOut", self.options.remove_transitions)
        self.attr_remove("autoText", self.options.remove_autoTexts)
        self.attr_remove("view", self.options.remove_views)


# Create effect instance.
if __name__ == "__main__":
    Uninstall().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2008 Stephen Silver
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
"""
Simple wrapper around ps2pdf
"""

import sys
import os

import inkex
from inkex.command import ProgramRunError, call, which
from inkex.localization import inkex_gettext as _


class PostscriptInput(inkex.CallExtension):
    """Load Postscript/EPS Files by calling ps2pdf program"""

    input_ext = "ps"
    output_ext = "pdf"
    multi_inx = True

    def add_arguments(self, pars):
        pars.add_argument("--crop", type=inkex.Boolean, default=False)
        pars.add_argument(
            "--autorotate", choices=["None", "PageByPage", "All"], default="None"
        )

    def call(self, input_file, output_file):
        crop = "-dEPSCrop" if self.options.crop else None
        if sys.platform == "win32":
            params = [
                "-q",
                "-P-",
                "-dSAFER",
                "-dNOPAUSE",
                "-dBATCH",
                "-sDEVICE#pdfwrite",
                "-dCompatibilityLevel#1.4",
                crop,
                "-sOutputFile#" + output_file,
                "-dAutoRotatePages#/" + self.options.autorotate,
                input_file,
            ]
            gs_execs = ["gswin64c", "gswin32c"]
            gs_exec = None
            for executable in gs_execs:
                try:
                    which(executable)
                    gs_exec = executable
                except:
                    pass
            if gs_exec is None:
                if "PYTEST_CURRENT_TEST" in os.environ:
                    gs_exec = "gswin64c"  # In CI, we have neither available,
                    # but there are mock files for the 64 bit version
                else:
                    raise inkex.AbortExtension(_("No GhostScript executable was found"))
            try:
                call(gs_exec, *params)
            except ProgramRunError as err:
                self.handle_gs_error(err)
        else:
            try:
                call(
                    "ps2pdf",
                    crop,
                    "-dAutoRotatePages=/" + self.options.autorotate,
                    input_file,
                    output_file,
                )
            except ProgramRunError as err:
                self.handle_gs_error(err)

    def handle_gs_error(self, err: ProgramRunError):
        inkex.errormsg(
            _(
                "Ghostscript was unable to read the file. \nThe following error message was returned:"
            )
            + "\n"
        )
        inkex.errormsg(err.stderr.decode("utf8"))
        inkex.errormsg(err.stdout.decode("utf8"))
        raise inkex.AbortExtension()


if __name__ == "__main__":
    PostscriptInput().run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Generic template functionality controlled by the INX file.
"""

import inkex


class InxDefinedTemplate(inkex.TemplateExtension):
    """Most functionality is in TemplateExtension"""

    multi_inx = True
    themes = {
        "white": ("#ffffff", "#666666"),
        "gray": ("#808080", "#444444"),
        "black": ("#000000", "#999999"),
    }

    def add_arguments(self, pars):
        pars.add_argument("--background", type=self.arg_theme, default="normal")
        pars.add_argument("--noborder", type=inkex.Boolean)

    def arg_theme(self, value):
        """Set the page theme based on the value given"""
        if value in self.themes:
            return self.themes[value]
        return None

    def set_namedview(self, width, height, unit):
        super(InxDefinedTemplate, self).set_namedview(width, height, unit)
        namedview = self.svg.namedview
        if self.options.background:
            namedview.set("pagecolor", self.options.background[0])
            namedview.set("bordercolor", self.options.background[1])
            namedview.set("inkscape:pageopacity", "1.0")
            namedview.set("inkscape:pageshadow", "0")

        if self.options.noborder:
            namedview.set("bordercolor", namedview.get("pagecolor"))
            namedview.set("borderopacity", "0")


if __name__ == "__main__":
    InxDefinedTemplate().run()



#
# Copyright (C) 2019-2021 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Searching for external packages and getting meta data about them.
"""

import re
import os
import json
import logging
import requests

try:
    from cachecontrol import CacheControl, CacheControlAdapter
    from cachecontrol.caches.file_cache import FileCache
    from cachecontrol.heuristics import ExpiresAfter
except (ImportError, ModuleNotFoundError):
    CacheControl = None

from inkex.command import CommandNotFound, ProgramRunError, call
from collections import defaultdict

from .utils import INKSCAPE_VERSION, CACHE_DIR
from .package import DEFAULT_ICON, PackageItem

PYTHON_VERSION = "py3"

class SearchError(IOError):
    pass

class LocalFile(object):
    """Same API as RemoteFile, but for locals"""

    def __init__(self, basedir, url):
        self.basedir = basedir
        self.url = url

    def __str__(self):
        return self.url

    def filename(self):
        return self.url.split("/")[-1]

    def filepath(self):
        return os.path.join(self.basedir, self.url)

    def as_local_file(self):
        filename = self.filepath()
        if os.path.isfile(filename):
            return filename
        raise IOError(f"Can't find file: {filename}")


class RemoteFile(object):
    """A remote file, icon, zip etc"""

    def __init__(self, session, url):
        self.session = session
        self.url = url

    def __str__(self):
        return self.url

    def get(self):
        return self.session.get(self.url)

    def filename(self):
        return self.url.split("/")[-1]

    def filepath(self):
        return os.path.join(CACHE_DIR, self.filename())

    def as_local_file(self):
        if self.url:
            remote = self.get()
            if remote and remote.status_code == 200:
                with open(self.filepath(), "wb") as fhl:
                    fhl.write(remote.content)
                return self.filepath()
        return None


class RemoteArchive(object):
    """The remote archive used to be PyPI but is now inkscape's own website."""

    URL = "https://inkscape.org/gallery/={category}/json/"

    def __init__(self, category, version=INKSCAPE_VERSION):
        self.version = version
        self.url = self.URL.format(category=category)
        self.session = requests.session()
        if CacheControl is not None:
            self.session.mount(
                "https://",
                CacheControlAdapter(
                    cache=FileCache(CACHE_DIR),
                    heuristic=ExpiresAfter(days=1),
                ),
            )

    def _remote_file(self, url):
        return RemoteFile(self.session, url)

    def search(self, query, filtered=True):
        """Search for extension packages"""
        for info in self._search(query):
            item = PackageItem(info, remote=self._remote_file)
            if not filtered or not self.version or self.version in item.targets:
                yield item

    def _search(self, query, tags=[]):
        """Search for the given query and yield each item, will raise if any other response"""
        response = None

        if query is None:
            try:
                response = self.session.get(
                    self.url,
                    params={
                        "tags": tags,
                        "checked": 1,
                        "limit": 10,
                        "order": "extra_status",
                    },
                )
            except requests.exceptions.RequestException as err:
                raise SearchError(str(err))
            except ConnectionError as err:
                raise SearchError(str(err))
            except requests.exceptions.RequestsWarning:
                pass
        else:
            response = self.session.get(
                self.url, params={"q": query, "tags": tags, "checked": 1}
            )
        if response is not None:
            for item in response.json()["items"]:
                yield item



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2012 Juan Pablo Carbajal ajuanpi-dev@gmail.com
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import math
import random
import inkex


class JitterNodes(inkex.EffectExtension):
    """Jiggle nodes around"""

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--radiusx", type=float, default=10.0, help="Randum radius X")
        pars.add_argument("--radiusy", type=float, default=10.0, help="Randum radius Y")
        pars.add_argument(
            "--ctrl", type=inkex.Boolean, default=False, help="Randomize ctrl points"
        )
        pars.add_argument(
            "--end", type=inkex.Boolean, default=True, help="Randomize nodes"
        )
        pars.add_argument(
            "--dist",
            type=self.arg_method("dist"),
            default=self.dist_uniform,
            help="Distribution of displacement",
        )

    def effect(self):
        for node in self.svg.selection.filter(inkex.PathElement):
            path = node.path.to_superpath()
            for subpath in path:
                closed = subpath[0] == subpath[-1]
                for index, csp in enumerate(subpath):
                    if closed and index == len(subpath) - 1:
                        subpath[index] = subpath[0]
                        break
                    if self.options.end:
                        delta = self.randomize([0, 0])
                        csp[0][0] += delta[0]
                        csp[0][1] += delta[1]
                        csp[1][0] += delta[0]
                        csp[1][1] += delta[1]
                        csp[2][0] += delta[0]
                        csp[2][1] += delta[1]
                    if self.options.ctrl:
                        csp[0] = self.randomize(csp[0])
                        csp[2] = self.randomize(csp[2])
            node.path = path

    def randomize(self, pos):
        """Randomise the given position [x, y] as set in the options"""
        delta = self.options.dist(self.options.radiusx, self.options.radiusy)
        return [pos[0] + delta[0], pos[1] + delta[1]]

    @staticmethod
    def dist_gaussian(x, y):
        """Gaussian distribution"""
        return random.gauss(0.0, x), random.gauss(0.0, y)

    @staticmethod
    def dist_pareto(x, y):
        """Pareto distribution"""
        # sign is used to fake a double sided pareto distribution.
        # for parameter value between 1 and 2 the distribution has infinite variance
        # I truncate the distribution to a high value and then normalize it.
        # The idea is to get spiky distributions, any distribution with long-tails is
        # good (ideal would be Levy distribution).
        sign = random.uniform(-1.0, 1.0)
        return x * math.copysign(
            min(random.paretovariate(1.0), 20.0) / 20.0, sign
        ), y * math.copysign(min(random.paretovariate(1.0), 20.0) / 20.0, sign)

    @staticmethod
    def dist_lognorm(x, y):
        """Log Norm distribution"""
        sign = random.uniform(-1.0, 1.0)
        return x * math.copysign(
            random.lognormvariate(0.0, 1.0) / 3.5, sign
        ), y * math.copysign(random.lognormvariate(0.0, 1.0) / 3.5, sign)

    @staticmethod
    def dist_uniform(x, y):
        """Uniform distribution"""
        return random.uniform(-x, x), random.uniform(-y, y)


if __name__ == "__main__":
    JitterNodes().run()



#!/bin/env python3
# coding=utf-8
#
# Copyright (C) 2019 Marc Jeanmougin, Cédric Gémy, a-l-e
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#


import os
import re
import sys
import inkex
from inkex import AbortExtension
from inkex.base import TempDirMixin
from inkex.command import take_snapshot, call
from inkex.localization import inkex_gettext as _


SCRIBUS_EXE = "scribus"
VERSION_REGEX = re.compile(r"(\d+)\.(\d+)\.(\d+)")


# several things could be taken into consideration here :
# - the fact that openDoc works on svg files is a workaround
# - the commented parts should be the correct way to do things
#   and even include a possibility to add margins
#   BUT currently fails to place the SVG document
#   (object placed top-left instead of SVG placed top-left)
class Scribus(TempDirMixin, inkex.OutputExtension):
    def add_arguments(self, arg_parser):
        arg_parser.add_argument(
            "--pdf-version",
            type=int,
            dest="pdfVersion",
            default=11,
            help="PDF version (e.g. integer numbers between 11 and 15, see Scribus documentation for details)",
        )
        arg_parser.add_argument(
            "--bleed", type=float, dest="bleed", default=0.0, help="Bleed value"
        )
        arg_parser.add_argument(
            "--bleed-marks",
            type=inkex.Boolean,
            dest="bleedMarks",
            default=False,
            help="Draw bleed marks",
        )
        arg_parser.add_argument(
            "--color-marks",
            type=inkex.Boolean,
            dest="colorMarks",
            default=False,
            help="Draw color Marks",
        )
        arg_parser.add_argument(
            "--intent",
            type=int,
            dest="intent",
            default=0,
            help="Rendering intent. Options: 0: Perceptual, 1: Relative Colorimetric, 2: Saturation, 3: Absolute Colorimetric",
        )
        arg_parser.add_argument(
            "--title",
            type=str,
            dest="title",
            default="",
            help="PDF title, required for PDF/X",
        )
        # arg_parser.add_argument("--fonts", type=int, dest="fonts", default="1",
        #                        help="Embed fonts : 0 for embedding, 1 to convert to path, 2 to prevent embedding")

    def generate_script(self, stream, width, height, icc):
        margin = self.options.bleed
        pdfVersion = self.options.pdfVersion
        embedFonts = 1  # self.options.fonts
        bleedMarks = self.options.bleedMarks
        colorMarks = self.options.colorMarks
        if (bleedMarks or colorMarks) and margin < 7:
            raise AbortExtension(
                _("You need at least 7mm bleed to show cutting marks or color marks")
            )
        if bleedMarks or colorMarks:
            margin = (
                margin - 7
            )  # because scribus is weird. At the time of 1.5.5, it adds 7 when those are set.
        stream.write(
            f"""
import scribus
import sys
icc = "{icc}"
margin = {margin}
class exportPDF():
    def __init__(self, svg=sys.argv[1], o=sys.argv[2]):
        #scribus.newDocument(({width},{height}), (margin,margin,margin,margin),
        #                    PORTRAIT, 1, UNIT_MILLIMETERS, PAGE_1, 0, 1)
        #scribus.placeSVG(svg, 0, 0)
        scribus.openDoc(svg)
        pdf = scribus.PDFfile()
        scribus.setUnit(UNIT_MILLIMETERS)
        pdf.bleedl = margin
        pdf.bleedr = margin
        pdf.bleedt = margin
        pdf.bleedb = margin
        pdf.useDocBleeds = False
        pdf.cropMarks = {bleedMarks}
        pdf.bleedMarks = {bleedMarks}
        pdf.colorMarks = {colorMarks}
        pdf.version = {pdfVersion}
        pdf.allowAnnots = True
        pdf.allowChange = True
        pdf.allowCopy = True
        pdf.allowPrinting = True
        pdf.noembicc = False #embed icc !
        pdf.solidpr = icc
        pdf.imagepr = icc
        pdf.printprofc = icc
        pdf.intenti = {self.options.intent}
        pdf.intents = {self.options.intent}
        pdf.info = "{self.options.title}"
        pdf.profiles = True
        pdf.profilei = True
        pdf.outdst = 1 # output destination : 0=screen, 1=printer
        pdf.file = o
        pdf.compress = True
        pdf.compressmtd = 0 # 0 = automatic, 1 = jpeg ; 2 = zip, 3 = none
        pdf.quality = 0 #max
        pdf.fontEmbedding = {embedFonts}
        pdf.thumbnails = True

        pdf.save()
exportPDF()"""
        )

    def save(self, stream):
        scribus_version = call(SCRIBUS_EXE, "-g", "--version")
        version_match = VERSION_REGEX.search(scribus_version)
        if version_match is None:
            raise AbortExtension(
                _("Could not detect Scribus version ()").format(scribus_version)
            )
        major = int(version_match.group(1))
        minor = int(version_match.group(2))
        point = int(version_match.group(3))
        if (major < 1) or (major == 1 and minor < 5):
            raise AbortExtension(
                _("Found Scribus {}. This extension requires Scribus 1.5.x.").format(
                    version_match.group(0)
                )
            )

        input_file = self.options.input_file
        py_file = os.path.join(self.tempdir, "scribus.py")
        svg_file = os.path.join(self.tempdir, "in.svg")
        profiles = self.svg.defs.findall("svg:color-profile")
        if len(profiles) == 0:
            raise AbortExtension(
                _("Please select a color profile in the document settings.")
            )
        elif len(profiles) > 1:
            raise AbortExtension(
                _(
                    "Please only link a single color profile in the document settings. "
                    "No output generated."
                )
            )
        iccPath = profiles[0].get("xlink:href")

        with open(input_file) as f:
            with open(svg_file, "w") as f1:
                for line in f:
                    f1.write(line)
            f.close()

        pdf_file = os.path.join(self.tempdir, "out.pdf")
        width = self.svg.unittouu(self.svg.get("width"))
        height = self.svg.unittouu(self.svg.get("height"))

        with open(py_file, "w") as fhl:
            self.generate_script(fhl, width, height, iccPath)
        call(SCRIBUS_EXE, "-g", "-py", py_file, svg_file, pdf_file)
        with open(pdf_file, "rb") as fhl:
            stream.write(fhl.read())


if __name__ == "__main__":
    Scribus().run()



#!/usr/bin/env python
#
# Copyright (C) 2016 su_v, <suv-sf@users.sf.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Convert mesh gradient to path
"""

import inkex
from inkex.elements import MeshGradient

# globals
EPSILON = 1e-3
MG_PROPS = ["fill", "stroke"]


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    """Test approximate equality.

    ref:
        PEP 485 -- A Function for testing approximate equality
        https://www.python.org/dev/peps/pep-0485/#proposed-implementation
    """
    # pylint: disable=invalid-name
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def reverse_path(csp):
    """Reverse path in CSP notation."""
    rcsp = []
    for subpath in reversed(csp):
        rsub = [list(reversed(cp)) for cp in reversed(subpath)]
        rcsp.append(rsub)
    return rcsp


def join_path(csp1, sp1, csp2, sp2):
    """Join sub-paths *sp1* and *sp2*."""
    pt1 = csp1[sp1][-1][1]
    pt2 = csp2[sp2][0][1]
    if isclose(pt1[0], pt2[0], EPSILON) and isclose(pt1[1], pt2[1], EPSILON):
        csp1[sp1][-1][2] = csp2[sp2][0][2]
        csp1[sp1].extend(csp2[sp2][1:])
    else:
        # inkex.debug('not close')
        csp1.append(csp2[sp2])
    return csp1


def is_url(val):
    """Check whether attribute value is linked resource."""
    return val.startswith("url(#")


def mesh_corners(meshgradient):
    """Return list of mesh patch corners, patch paths."""
    rows = len(meshgradient)
    cols = len(meshgradient[0])
    # first corner of mesh gradient
    corner_x = float(meshgradient.get("x", "0.0"))
    corner_y = float(meshgradient.get("y", "0.0"))
    # init corner and meshpatch lists
    corners = [[None for _ in range(cols + 1)] for _ in range(rows + 1)]
    corners[0][0] = [corner_x, corner_y]
    meshpatch_csps = []
    for meshrow in range(rows):
        for meshpatch in range(cols):
            # get start point for current meshpatch edges
            if meshrow == 0:
                first_corner = corners[meshrow][meshpatch]
            if meshrow > 0:
                first_corner = corners[meshrow][meshpatch + 1]
            # parse path of meshpatch edges
            path = "M {},{}".format(*first_corner)
            for edge in meshgradient[meshrow][meshpatch]:
                path = " ".join([path, edge.get("path")])
            csp = inkex.Path(path).to_superpath()
            # update corner list with current meshpatch
            if meshrow == 0:
                corners[meshrow][meshpatch + 1] = csp[0][1][1]
                corners[meshrow + 1][meshpatch + 1] = csp[0][2][1]
                if meshpatch == 0:
                    corners[meshrow + 1][meshpatch] = csp[0][3][1]
            if meshrow > 0:
                corners[meshrow][meshpatch + 1] = csp[0][0][1]
                corners[meshrow + 1][meshpatch + 1] = csp[0][1][1]
                if meshpatch == 0:
                    corners[meshrow + 1][meshpatch] = csp[0][2][1]
            # append to list of meshpatch csp
            meshpatch_csps.append(csp)
    return corners, meshpatch_csps


def mesh_hvlines(meshgradient):
    """Return lists of vertical and horizontal patch edges."""
    rows = len(meshgradient)
    cols = len(meshgradient[0])
    # init lists for horizontal, vertical lines
    hlines = [[None for _ in range(cols)] for _ in range(rows + 1)]
    vlines = [[None for _ in range(rows)] for _ in range(cols + 1)]
    for meshrow in range(rows):
        for meshpatch in range(cols):
            # horizontal edges
            if meshrow == 0:
                edge = meshgradient[meshrow][meshpatch][0]
                hlines[meshrow][meshpatch] = edge.get("path")
                edge = meshgradient[meshrow][meshpatch][2]
                hlines[meshrow + 1][meshpatch] = edge.get("path")
            if meshrow > 0:
                edge = meshgradient[meshrow][meshpatch][1]
                hlines[meshrow + 1][meshpatch] = edge.get("path")
            # vertical edges
            if meshrow == 0:
                edge = meshgradient[meshrow][meshpatch][1]
                vlines[meshpatch + 1][meshrow] = edge.get("path")
                if meshpatch == 0:
                    edge = meshgradient[meshrow][meshpatch][3]
                    vlines[meshpatch][meshrow] = edge.get("path")
            if meshrow > 0:
                edge = meshgradient[meshrow][meshpatch][0]
                vlines[meshpatch + 1][meshrow] = edge.get("path")
                if meshpatch == 0:
                    edge = meshgradient[meshrow][meshpatch][2]
                    vlines[meshpatch][meshrow] = edge.get("path")
    return hlines, vlines


def mesh_to_outline(corners, hlines, vlines):
    """Construct mesh outline as CSP path."""
    outline_csps = []
    path = "M {},{}".format(*corners[0][0])
    for edge_path in hlines[0]:
        path = " ".join([path, edge_path])
    for edge_path in vlines[-1]:
        path = " ".join([path, edge_path])
    for edge_path in reversed(hlines[-1]):
        path = " ".join([path, edge_path])
    for edge_path in reversed(vlines[0]):
        path = " ".join([path, edge_path])
    outline_csps.append(inkex.Path(path).to_superpath())
    return outline_csps


def mesh_to_grid(corners, hlines, vlines):
    """Construct mesh grid with CSP paths."""
    rows = len(corners) - 1
    cols = len(corners[0]) - 1
    gridline_csps = []
    # horizontal
    path = "M {},{}".format(*corners[0][0])
    for edge_path in hlines[0]:
        path = " ".join([path, edge_path])
    gridline_csps.append(inkex.Path(path).to_superpath())
    for i in range(1, rows + 1):
        path = "M {},{}".format(*corners[i][-1])
        for edge_path in reversed(hlines[i]):
            path = " ".join([path, edge_path])
        gridline_csps.append(inkex.Path(path).to_superpath())
    # vertical
    path = "M {},{}".format(*corners[-1][0])
    for edge_path in reversed(vlines[0]):
        path = " ".join([path, edge_path])
    gridline_csps.append(inkex.Path(path).to_superpath())
    for j in range(1, cols + 1):
        path = "M {},{}".format(*corners[0][j])
        for edge_path in vlines[j]:
            path = " ".join([path, edge_path])
        gridline_csps.append(inkex.Path(path).to_superpath())
    return gridline_csps


def mesh_to_faces(corners, hlines, vlines):
    """Construct mesh faces with CSP paths."""
    rows = len(corners) - 1
    cols = len(corners[0]) - 1
    face_csps = []
    for row in range(rows):
        for col in range(cols):
            # init new face
            face = []
            # init edge paths
            edge_t = hlines[row][col]
            edge_b = hlines[row + 1][col]
            edge_l = vlines[col][row]
            edge_r = vlines[col + 1][row]
            # top edge, first
            if row == 0:
                path = "M {},{}".format(*corners[row][col])
                path = " ".join([path, edge_t])
                face.append(inkex.Path(path).to_superpath()[0])
            else:
                path = "M {},{}".format(*corners[row][col + 1])
                path = " ".join([path, edge_t])
                face.append(reverse_path(inkex.Path(path).to_superpath())[0])
            # right edge
            path = "M {},{}".format(*corners[row][col + 1])
            path = " ".join([path, edge_r])
            join_path(face, -1, inkex.Path(path).to_superpath(), 0)
            # bottom edge
            path = "M {},{}".format(*corners[row + 1][col + 1])
            path = " ".join([path, edge_b])
            join_path(face, -1, inkex.Path(path).to_superpath(), 0)
            # left edge
            if col == 0:
                path = "M {},{}".format(*corners[row + 1][col])
                path = " ".join([path, edge_l])
                join_path(face, -1, inkex.Path(path).to_superpath(), 0)
            else:
                path = "M {},{}".format(*corners[row][col])
                path = " ".join([path, edge_l])
                join_path(face, -1, reverse_path(inkex.Path(path).to_superpath()), 0)
            # append face to output list
            face_csps.append(face)
    return face_csps


class MeshToPath(inkex.EffectExtension):
    """Effect extension to convert mesh geometry to path data."""

    def add_arguments(self, pars):
        pars.add_argument("--tab", help="The selected UI-tab")
        pars.add_argument("--mode", default="outline", help="Edge mode")

    def process_props(self, mdict, res_type="meshgradient"):
        """Process style properties of style dict *mdict*."""
        result = []
        for key, val in mdict.items():
            if key in MG_PROPS:
                if is_url(val):
                    paint_server = self.svg.getElementById(val)
                    if res_type == "meshgradient" and isinstance(
                        paint_server, MeshGradient
                    ):
                        result.append(paint_server)
        return result

    def process_style(self, node, res_type="meshgradient"):
        """Process style of *node*."""
        result = node.specified_style()
        result = self.process_props(result, res_type)
        # TODO: check for child paint servers
        return result

    def find_meshgradients(self, node):
        """Parse node style, return list with linked meshgradients."""
        return self.process_style(node, res_type="meshgradient")

    # ----- Process meshgradient definitions

    def mesh_to_csp(self, meshgradient):
        """Parse mesh geometry and build csp-based path data."""

        # init variables
        transform = None
        mode = self.options.mode

        # gradient units
        mesh_units = meshgradient.get("gradientUnits", "objectBoundingBox")
        if mesh_units == "objectBoundingBox":
            # TODO: position and scale based on "objectBoundingBox" units
            return

        # Inkscape SVG 0.92 and SVG 2.0 draft mesh transformations
        transform = meshgradient.gradientTransform @ meshgradient.transform

        # parse meshpatches, calculate absolute corner coords
        corners, meshpatch_csps = mesh_corners(meshgradient)

        if mode == "meshpatches":
            return meshpatch_csps, transform
        else:
            hlines, vlines = mesh_hvlines(meshgradient)
            if mode == "outline":
                return mesh_to_outline(corners, hlines, vlines), transform
            elif mode == "gridlines":
                return mesh_to_grid(corners, hlines, vlines), transform
            elif mode == "faces":
                return mesh_to_faces(corners, hlines, vlines), transform

    # ----- Convert meshgradient definitions

    def csp_to_path(self, node, csp_list, transform=None):
        """Create new paths based on csp data, return group with paths."""
        # set up stroke width, group
        stroke_width = self.svg.viewport_to_unit("1px")
        stroke_color = "#000000"
        style = {
            "fill": "none",
            "stroke": stroke_color,
            "stroke-width": str(stroke_width),
        }

        group = inkex.Group()
        # apply gradientTransform and node's preserved transform to group
        group.transform = transform @ node.transform

        # convert each csp to path, append to group
        for csp in csp_list:
            elem = group.add(inkex.PathElement())
            elem.style = style
            elem.path = inkex.CubicSuperPath(csp)
            if self.options.mode == "outline":
                elem.path.close()
            elif self.options.mode == "faces":
                if len(csp) == 1 and len(csp[0]) == 5:
                    elem.path.close()
        return group

    def effect(self):
        """Main routine to convert mesh geometry to path data."""
        # loop through selection
        for node in self.svg.selection.values():
            meshgradients = self.find_meshgradients(node)
            # if style references meshgradient
            if meshgradients:
                for meshgradient in meshgradients:
                    csp_list = None
                    result = None
                    # parse mesh geometry
                    if meshgradient is not None:
                        csp_list, mat = self.mesh_to_csp(meshgradient)
                    # generate new paths with path data based on mesh geometry
                    if csp_list is not None:
                        result = self.csp_to_path(node, csp_list, mat)
                    # add result (group) to document
                    if result is not None:
                        index = node.getparent().index(node)
                        node.getparent().insert(index + 1, result)


if __name__ == "__main__":
    MeshToPath().run()



#!/usr/bin/env python
"""Remove green color from selected objects"""

import inkex


class RemoveGreen(inkex.ColorExtension):
    """Remove green color from selected objects"""

    def modify_color(self, name, color):
        return inkex.Color([color.red, 0, color.blue])


if __name__ == "__main__":
    RemoveGreen().run()



# coding=utf-8
#
# Copyright (C) 2021 - Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Toplevel CSS utils that do not depend on other inkex functionality

.. versionadded:: 1.2
    Previously a part of :py:mod:`inkex.styles`"""


import re
import cssselect


class ConditionalRule:
    """A single css rule

    .. versionchanged:: 1.2
        The CSS rule is now processed using cssselect."""

    step_to_xpath = [
        # namespace addition
        (re.compile(r"(::|\/)([a-z]+)(?=\W)(?!-)"), r"\1svg:\2"),
    ]

    def __init__(self, rule):
        self.rule = rule.strip()
        self.selector = cssselect.parse(self.rule)[0]

    def __str__(self):
        return self.rule

    def to_xpath(self):
        """Attempt to convert the rule into a simplified xpath"""
        # the space in the end is needed for the negative lookbehind in the regex, will
        # be removed on return
        ret = cssselect.HTMLTranslator().selector_to_xpath(self.selector) + " "
        for matcher, replacer in self.step_to_xpath:
            ret = matcher.sub(replacer, ret)
        return ret.strip()

    def get_specificity(self):
        """gets the css specificity of this selector

        .. versionadded:: 1.2"""
        return self.selector.specificity()



# coding=utf-8
#
# Copyright (C) 2018 - Martin Owens <doctormo@mgail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Provide some documentation to existing extensions about why they're failing.
"""
#
# We ignore a lot of pylint warnings here:
#
# pylint: disable=invalid-name,unused-argument,missing-docstring,too-many-public-methods
#

import os
import sys
import warnings
import argparse

from ..transforms import Transform
from .. import utils
from .. import units
from ..elements._base import BaseElement, ShapeElement
from ..elements._selected import ElementList
from .meta import deprecate, _deprecated

warnings.simplefilter("default")
# To load each of the deprecated sub-modules (the ones without a namespace)
# we will add the directory to our pythonpath so older scripts can find them

INKEX_DIR = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
SIMPLE_DIR = os.path.join(INKEX_DIR, "deprecated-simple")

if os.path.isdir(SIMPLE_DIR):
    sys.path.append(SIMPLE_DIR)


class DeprecatedDict(dict):
    @deprecate
    def __getitem__(self, key):
        return super().__getitem__(key)

    @deprecate
    def __iter__(self):
        return super().__iter__()


# legacy inkex members


class lazyproxy:
    """Proxy, use as decorator on a function with provides the wrapped object.
    The decorated function is called when a member is accessed on the proxy.
    """

    def __init__(self, getwrapped):
        """
        :param getwrapped: Callable which returns the wrapped object
        """
        self._getwrapped = getwrapped

    def __getattr__(self, name):
        return getattr(self._getwrapped(), name)

    def __call__(self, *args, **kwargs):
        return self._getwrapped()(*args, **kwargs)


@lazyproxy
def localize():
    _deprecated("inkex.localize was moved to inkex.localization.localize.", stack=3)
    from ..localization import localize as wrapped

    return wrapped


def are_near_relative(a, b, eps):
    _deprecated(
        "inkex.are_near_relative was moved to " "inkex.units.are_near_relative", stack=2
    )
    return units.are_near_relative(a, b, eps)


def debug(what):
    _deprecated("inkex.debug was moved to inkex.utils.debug.", stack=2)
    return utils.debug(what)


# legacy inkex members <= 0.48.x


def unittouu(string):
    _deprecated(
        "inkex.unittouu is now a method in the SvgDocumentElement class. "
        "Use `self.svg.unittouu(str)` instead.",
        stack=2,
    )
    return units.convert_unit(string, "px")


# optparse.Values.ensure_value


def ensure_value(self, attr, value):
    _deprecated("Effect().options.ensure_value was removed.", stack=2)
    if getattr(self, attr, None) is None:
        setattr(self, attr, value)
    return getattr(self, attr)


argparse.Namespace.ensure_value = ensure_value  # type: ignore


@deprecate
def zSort(inNode, idList):
    """self.svg.get_z_selected()"""
    sortedList = []
    theid = inNode.get("id")
    if theid in idList:
        sortedList.append(theid)
    for child in inNode:
        if len(sortedList) == len(idList):
            break
        sortedList += zSort(child, idList)
    return sortedList


# This can't be handled as a mixin class because of circular importing.
def description(self, value):
    """Use elem.desc = value"""
    self.desc = value


BaseElement.description = deprecate(description, "1.1")


def composed_style(element: ShapeElement):
    """Calculate the final styles applied to this element
    This function has been deprecated in favor of BaseElement.specified_style()"""
    return element.specified_style()


ShapeElement.composed_style = deprecate(composed_style, "1.2")


def paint_order(selection: ElementList):
    """Use :func:`rendering_order`"""
    return selection.rendering_order()


ElementList.paint_order = deprecate(paint_order, "1.2")  # type: ignore


def transform_imul(self, matrix):
    """Use @= operator instead"""
    return self.__imatmul__(matrix)


def transform_mul(self, matrix):
    """Use @ operator instead"""
    return self.__matmul__(matrix)


Transform.__imul__ = deprecate(transform_imul, "1.2")  # type: ignore
Transform.__mul__ = deprecate(transform_mul, "1.2")  # type: ignore



#
# Copyright 2018 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""Gtk 3+ Inkscape 1.0 Extensions Manager GUI."""

import os
import sys
import webbrowser

from inkex.gui import Window, ChildWindow, TreeView, asyncme
from inkex.gui.pixmap import PixmapManager
from inkex.gui.app import Gtk

from ..utils import DATA_DIR, INKSCAPE_PROFILE
from ..remote import RemoteArchive, SearchError
from .. import __state__


def tag(name, color="#EFEFEF", fgcolor="#333333"):
    return f"<span size='small' font_family='monospace' style='italic' background='{color}' foreground='{fgcolor}'> {name} </span> "


class LocalTreeView(TreeView):
    """A List of locally installed packages"""

    def __init__(self, target, widget, *args, **kwargs):
        self._target = target
        self._pixmaps = kwargs["pixmaps"]
        self._version = kwargs.pop("version", None)

        style = widget.get_style_context()
        self._fg_color = style.get_color(Gtk.StateFlags.NORMAL)
        # self._bg_color = style.get_property('background-color', Gtk.StateFlags.NORMAL)

        super().__init__(widget, *args, **kwargs)

    def get_name(self, item):
        # color = self._style.get_property('color', Gtk.StateFlags.NORMAL)
        return f"<big><b>{item.name}</b></big>\n<i>{item.summary}</i>"

    def get_icon(self, item):
        try:
            return self._pixmaps.get(item.get_icon(), "")
        except Exception:
            return ""

    def setup(self, *args, **kwargs):
        """Setup the treeview with one or many columns manually"""

        col = self.create_column("Extensions Package", expand=True)

        img = col.add_image_renderer(self.get_icon, pad=0, pixmaps=self._pixmaps)
        img.set_property("ypad", 2)

        txt = col.add_text_renderer(self.get_name)
        txt.set_property("foreground-rgba", self._fg_color)
        # This is detected as black for remote treeview, we're not sure why
        # txt.set_property("background-rgba", self._bg_color)
        txt.set_property("xpad", 8)

        col = self.create_column("Version", expand=False)
        txt = col.add_text_renderer("version", template=f"<i>{0}</i>")
        txt.set_property("foreground-rgba", self._fg_color)

        col = self.create_column("Author", expand=False)
        txt = col.add_text_renderer("author")
        txt.set_property("foreground-rgba", self._fg_color)

        if self._target.version_specific:
            col = self.create_column("Inkscape", expand=False)
            txt = col.add_text_renderer("target")
            txt.set_property("foreground-rgba", self._fg_color)

        self.create_sort(data=lambda item: item.name, ascending=True, contains="name")

        super().setup(*args, **kwargs)


class RemoteTreeView(LocalTreeView):
    """A List of remote packages for installation"""

    def get_installed(self, item):
        if item.installed is None:
            item.set_installed(False)
            for installed_item in self._target.list_installed():
                if item.ident == installed_item.ident:
                    item.set_installed(True)
        return bool(item.installed)

    def setup(self, *args, **kwargs):
        def get_star(item):
            star = "star-none.svg"
            if item.stars > 10:
                star = "star-lots.svg"
            elif item.stars > 1:
                star = "star-some.svg"
            return self._pixmaps.get(star)

        def get_sort(item):
            if not item.verified:
                return -100 + item.stars
            if self._version not in item.targets:
                return -20 + item.stars
            return item.stars + (1000 * int(item.recommended))

        super().setup(*args, **kwargs)

        col = self.create_column("Stars", expand=True)
        img = col.add_image_renderer(get_star, pad=0, size=16, pixmaps=self._pixmaps)
        img.set_property("ypad", 2)
        txt = col.add_text_renderer("stars")
        txt.set_property("foreground-rgba", self._fg_color)

        self.create_sort(data=get_sort, ascending=True)

    def get_name(self, item):
        summary = item.summary
        if self.get_installed(item):
            summary = tag("Installed", "#37943e", "white") + summary.strip()
        elif not item.verified:
            summary = tag("Unverified", "#CB0000", "white") + summary.strip()
        elif self._version not in item.targets:
            targets = item.targets or ["No Inkscape Versions"]
            summary = (
                "".join([tag(t, "#00CB00", "white") for t in targets]) + summary.strip()
            )
        else:
            summary = (
                tag("Verified", "#37943e", "white")
                + tag(self._version, "#37943e", "white")
                + summary.strip()
            )

        return f"<big><b>{item.name}</b></big>\n<i>{summary}</i>\n" + "".join(
            tag(t) for t in item.tags
        )


class LocalPixmaps(PixmapManager):
    """Load local pictures"""

    pixmap_dir = DATA_DIR


class ExtensionManagerWindow(Window):
    name = "gui"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target = self.gapp.kwargs["target"]
        self._version = self.gapp.kwargs["version"]

        self.pixmaps = LocalPixmaps("pixmaps", size=48, load_size=(96, 96))
        self.searching = self.widget("dl-searching")
        self.searchbox = self.searching.get_parent()
        self.searchbox.remove(self.searching)

        self.local = LocalTreeView(
            self.target,
            self.widget("local_extensions"),
            version=self._version,
            pixmaps=self.pixmaps,
            selected=self.select_local,
        )

        self.remote = RemoteTreeView(
            self.target,
            self.widget("remote_extensions"),
            version=self._version,
            pixmaps=self.pixmaps,
            selected=self.select_remote,
        )

        self.widget("loading").start()
        self.window.show_all()
        self.window.present()
        # self.window.set_keep_above(True)

        self.refresh_local()
        self.widget("remote_install").set_sensitive(False)
        self.widget("remote_info").set_sensitive(False)
        self.widget("local_install").set_sensitive(True)
        self.window.set_title(self.target.label + f" ({self._version}) - {__state__}")

        if not self.target.is_search:
            self.widget("remote_getlist").show()
            self.widget("dl-search").hide()
        else:
            self.widget("dl-search").show()
            self.widget("remote_getlist").hide()
        self.widget("no_results").hide()

        # Get a tasting of items from the server
        self._remote_search(None)

    def remote_getlist(self, widget):
        self._remote_search("")

    @asyncme.run_or_none
    def refresh_local(self):
        """Searches for locally installed extensions and adds them"""
        self.local.clear()
        self.widget("local_uninstall").set_sensitive(False)
        self.widget("local_information").set_sensitive(False)

        all_packages = []
        for item in self.target.list_installed(cached=False):
            self.local.add_item(item)
        self.widget("loading").stop()

    def select_local(self, item):
        """Select an installed extension"""
        self.widget("local_uninstall").set_sensitive(item.is_uninstallable())
        self.widget("local_information").set_sensitive(bool(item))

    def local_information(self, widget):
        """Show the more information window"""
        if self.local.selected:
            self.load_window("info", pixmaps=self.pixmaps, item=self.local.selected)

    def local_uninstall(self, widget):
        """Uninstall selected extection package"""
        item = self.local.selected
        if item.is_uninstallable():
            if item.uninstall():
                self.local.remove_item(item)
                self.remote.refresh()

    def change_remote_all(self, widget, unk):
        """When the source switch button is clicked"""
        self.remote_search(self.widget("dl-search"))

    def remote_search(self, widget):
        """Remote search activation"""
        query = widget.get_text()
        if len(query) > 2:
            self._remote_search(query)

    def _remote_search(self, query):
        filtered = False  # self.widget('remote_target').get_active()
        self.remote.clear()
        self.widget("no_results").hide()
        self.widget("results_found").show()
        self.widget("remote_install").set_sensitive(False)
        self.widget("remote_info").set_sensitive(False)
        self.widget("dl-search").set_sensitive(False)
        self.searchbox.add(self.searching)
        self.widget("dl-searching").start()
        self.async_search(query, filtered)

    def remote_info(self, widget):
        """Show the remote information"""
        if self.remote.selected and self.remote.selected.link:
            webbrowser.open(self.remote.selected.link)

    @asyncme.run_or_none
    def async_search(self, query, filtered):
        """Asyncronous searching in PyPI"""
        try:
            for package in self.target.search(query, filtered):
                self.add_search_result(package)
        except SearchError as err:
            # Ignore taster request
            if query is not None:
                self.dialog(str(err))
        self.search_finished()

    @asyncme.mainloop_only
    def add_search_result(self, package):
        """Adding things to Gtk must be done in mainloop"""
        self.remote.add_item(package)

    @asyncme.mainloop_only
    def search_finished(self):
        """After everything, finish the search"""
        self.searchbox.remove(self.searching)
        self.widget("dl-search").set_sensitive(True)
        self.replace(self.searching, self.remote._list)
        if not self.remote._model.iter_n_children():
            self.widget("no_results").show()
            self.widget("results_found").hide()

    def select_remote(self, item):
        """Select an installed extension"""
        # Do we have a place to install packages to?
        active = (
            self.remote.selected
            and self.remote.selected.is_installable(self._version)
            and not self.remote.selected.installed
        )
        self.widget("remote_install").set_sensitive(active)
        if self.remote.selected.link:
            self.widget("remote_info").set_sensitive(True)

    def remote_install(self, widget):
        """Install a remote package"""
        self.widget("remote_install").set_sensitive(False)
        item = self.remote.selected
        try:
            item.install()
        except Exception as err:
            self.dialog(str(err))
            return
        self.remote.refresh()
        self.refresh_local()
        self.widget("main_notebook").set_current_page(0)

    def local_install(self, widget):
        """Install from a local filename"""
        dialog = Gtk.FileChooserDialog(
            title="Please choose a package file",
            transient_for=self.window,
            action=Gtk.FileChooserAction.OPEN,
        )

        dialog.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OPEN,
            Gtk.ResponseType.OK,
        )

        filter_py = Gtk.FileFilter()
        filter_py.set_name("Packages")
        filter_py.add_mime_type("application/x-compressed-tar")
        filter_py.add_mime_type("application/zip")
        dialog.add_filter(filter_py)

        filter_text = Gtk.FileFilter()
        filter_text.set_name("Python Wheel")
        filter_text.add_mime_type("application/zip")
        filter_text.add_pattern("*.whl")
        dialog.add_filter(filter_text)

        filter_any = Gtk.FileFilter()
        filter_any.set_name("Any files")
        filter_any.add_pattern("*")
        dialog.add_filter(filter_any)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            self.target._install(dialog.get_filename(), {})
            self.refresh_local()

        dialog.destroy()

    def update_all(self, widget=None):
        """Update the extensions manager, and everything else"""
        pass

    def dialog(self, msg):
        self.widget("dialog_msg").set_label(msg)
        self.widget("dialog").set_transient_for(self.window)
        self.widget("dialog").show_all()

    def close_dialog(self, widget):
        self.widget("dialog_msg").set_label("")
        self.widget("dialog").hide()



#!/usr/bin/env python
"""Add more hue"""

import inkex


class MoreHue(inkex.ColorExtension):
    """Add hue to any selected object"""

    def modify_color(self, name, color):
        color.hue += int(0.05 * 255.0)
        return color


if __name__ == "__main__":
    MoreHue().run()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Sergei Izmailov <sergei.a.izmailov@gmail.com>
#                    Thomas Holder <thomas.holder@schrodinger.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=arguments-differ
"""
Element interface for patterns, filters, gradients and path effects.
"""

from typing import List, Tuple, TYPE_CHECKING

from lxml import etree

from ..transforms import Transform
from ..utils import parse_percent

from ..styles import Style

from ._utils import addNS
from ._base import BaseElement


if TYPE_CHECKING:
    from ._svg import SvgDocumentElement


class Filter(BaseElement):
    """A filter (usually in defs)"""

    tag_name = "filter"

    def add_primitive(self, fe_type, **args):
        """Create a filter primitive with the given arguments"""
        elem = etree.SubElement(self, addNS(fe_type, "svg"))
        elem.update(**args)
        return elem

    class Primitive(BaseElement):
        """Any filter primitive"""

    class Blend(Primitive):
        """Blend Filter element"""

        tag_name = "feBlend"

    class ColorMatrix(Primitive):
        """ColorMatrix Filter element"""

        tag_name = "feColorMatrix"

    class ComponentTransfer(Primitive):
        """ComponentTransfer Filter element"""

        tag_name = "feComponentTransfer"

    class Composite(Primitive):
        """Composite Filter element"""

        tag_name = "feComposite"

    class ConvolveMatrix(Primitive):
        """ConvolveMatrix Filter element"""

        tag_name = "feConvolveMatrix"

    class DiffuseLighting(Primitive):
        """DiffuseLightning Filter element"""

        tag_name = "feDiffuseLighting"

    class DisplacementMap(Primitive):
        """Flood Filter element"""

        tag_name = "feDisplacementMap"

    class Flood(Primitive):
        """DiffuseLightning Filter element"""

        tag_name = "feFlood"

    class GaussianBlur(Primitive):
        """GaussianBlur Filter element"""

        tag_name = "feGaussianBlur"

    class Image(Primitive):
        """Image Filter element"""

        tag_name = "feImage"

    class Merge(Primitive):
        """Merge Filter element"""

        tag_name = "feMerge"

    class Morphology(Primitive):
        """Morphology Filter element"""

        tag_name = "feMorphology"

    class Offset(Primitive):
        """Offset Filter element"""

        tag_name = "feOffset"

    class SpecularLighting(Primitive):
        """SpecularLighting Filter element"""

        tag_name = "feSpecularLighting"

    class Tile(Primitive):
        """Tile Filter element"""

        tag_name = "feTile"

    class Turbulence(Primitive):
        """Turbulence Filter element"""

        tag_name = "feTurbulence"


class Stop(BaseElement):
    """Gradient stop

    .. versionadded:: 1.1"""

    tag_name = "stop"

    @property
    def offset(self) -> float:
        """The offset of the gradient stop"""
        return self.get("offset")

    @offset.setter
    def offset(self, number):
        self.set("offset", number)

    def interpolate(self, other, fraction):
        """Interpolate gradient stops"""
        from ..tween import StopInterpolator

        return StopInterpolator(self, other).interpolate(fraction)


class Pattern(BaseElement):
    """Pattern element which is used in the def to control repeating fills"""

    tag_name = "pattern"
    WRAPPED_ATTRS = BaseElement.WRAPPED_ATTRS + (("patternTransform", Transform),)


class Gradient(BaseElement):
    """A gradient instruction usually in the defs."""

    WRAPPED_ATTRS = BaseElement.WRAPPED_ATTRS + (("gradientTransform", Transform),)
    """Additional to the :attr:`~inkex.elements._base.BaseElement.WRAPPED_ATTRS` of 
    :class:`~inkex.elements._base.BaseElement`, ``gradientTransform`` is wrapped."""

    orientation_attributes = ()  # type: Tuple[str, ...]
    """
    .. versionadded:: 1.1
    """

    @property
    def stops(self):
        """Return an ordered list of own or linked stop nodes

        .. versionadded:: 1.1"""
        gradcolor = (
            self.href
            if isinstance(self.href, (LinearGradient, RadialGradient))
            else self
        )
        return sorted(
            [child for child in gradcolor if isinstance(child, Stop)],
            key=lambda x: parse_percent(x.offset),
        )

    @property
    def stop_offsets(self):
        # type: () -> List[float]
        """Return a list of own or linked stop offsets

        .. versionadded:: 1.1"""
        return [child.offset for child in self.stops]

    @property
    def stop_styles(self):  # type: () -> List[Style]
        """Return a list of own or linked offset styles

        .. versionadded:: 1.1"""
        return [child.style for child in self.stops]

    def remove_orientation(self):
        """Remove all orientation attributes from this element

        .. versionadded:: 1.1"""
        for attr in self.orientation_attributes:
            self.pop(attr)

    def interpolate(
        self, other, fraction, svg=None
    ):  # type: (LinearGradient, float, SvgDocumentElement) -> LinearGradient
        """Interpolate with another gradient.

        .. versionadded:: 1.1"""
        from ..tween import GradientInterpolator

        return GradientInterpolator(self, other, svg).interpolate(fraction)

    def stops_and_orientation(self):
        """Return a copy of all the stops in this gradient

        .. versionadded:: 1.1"""
        stops = self.copy()
        stops.remove_orientation()
        orientation = self.copy()
        orientation.remove_all(Stop)
        return stops, orientation


class LinearGradient(Gradient):
    """LinearGradient element"""

    tag_name = "linearGradient"
    orientation_attributes = ("x1", "y1", "x2", "y2")
    """
    .. versionadded:: 1.1
    """

    def apply_transform(self):  # type: () -> None
        """Apply transform to orientation points and set it to identity.
        .. versionadded:: 1.1
        """
        trans = self.pop("gradientTransform")
        pt1 = (
            self.to_dimensionless(self.get("x1")),
            self.to_dimensionless(self.get("y1")),
        )
        pt2 = (
            self.to_dimensionless(self.get("x2")),
            self.to_dimensionless(self.get("y2")),
        )
        p1t = trans.apply_to_point(pt1)
        p2t = trans.apply_to_point(pt2)
        self.update(
            x1=self.to_dimensionless(p1t[0]),
            y1=self.to_dimensionless(p1t[1]),
            x2=self.to_dimensionless(p2t[0]),
            y2=self.to_dimensionless(p2t[1]),
        )


class RadialGradient(Gradient):
    """RadialGradient element"""

    tag_name = "radialGradient"
    orientation_attributes = ("cx", "cy", "fx", "fy", "r")
    """
    .. versionadded:: 1.1
    """

    def apply_transform(self):  # type: () -> None
        """Apply transform to orientation points and set it to identity.

        .. versionadded:: 1.1
        """
        trans = self.pop("gradientTransform")
        pt1 = (
            self.to_dimensionless(self.get("cx")),
            self.to_dimensionless(self.get("cy")),
        )
        pt2 = (
            self.to_dimensionless(self.get("fx")),
            self.to_dimensionless(self.get("fy")),
        )
        p1t = trans.apply_to_point(pt1)
        p2t = trans.apply_to_point(pt2)
        self.update(
            cx=self.to_dimensionless(p1t[0]),
            cy=self.to_dimensionless(p1t[1]),
            fx=self.to_dimensionless(p2t[0]),
            fy=self.to_dimensionless(p2t[1]),
        )


class PathEffect(BaseElement):
    """Inkscape LPE element"""

    tag_name = "inkscape:path-effect"


class MeshGradient(Gradient):
    """Usable MeshGradient XML base class

    .. versionadded:: 1.1"""

    tag_name = "meshgradient"

    @classmethod
    def new_mesh(cls, pos=None, rows=1, cols=1, autocollect=True):
        """Return skeleton of 1x1 meshgradient definition."""
        # initial point
        if pos is None or len(pos) != 2:
            pos = [0.0, 0.0]
        # create nested elements for rows x cols mesh
        meshgradient = cls()
        for _ in range(rows):
            meshrow: BaseElement = meshgradient.add(MeshRow())
            for _ in range(cols):
                meshrow.append(MeshPatch())
        # set meshgradient attributes
        meshgradient.set("gradientUnits", "userSpaceOnUse")
        meshgradient.set("x", pos[0])
        meshgradient.set("y", pos[1])
        if autocollect:
            meshgradient.set("inkscape:collect", "always")
        return meshgradient


class MeshRow(BaseElement):
    """Each row of a mesh gradient

    .. versionadded:: 1.1"""

    tag_name = "meshrow"


class MeshPatch(BaseElement):
    """Each column or 'patch' in a mesh gradient

    .. versionadded:: 1.1"""

    tag_name = "meshpatch"

    def stops(self, edges, colors):
        """Add or edit meshpatch stops with path and stop-color."""
        # iterate stops based on number of edges (path data)
        for i, edge in enumerate(edges):
            if i < len(self):
                stop = self[i]
            else:
                stop = self.add(Stop())

            # set edge path data
            stop.set("path", str(edge))
            # set stop color
            stop.style["stop-color"] = str(colors[i % 2])



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2011 Vincent Nivoliers and contributors
#
# Contributors
#    ~suv, <suv-sf@users.sf.net>
# - Voronoi Diagram algorithm and C code by Steven Fortune, 1987, http://ect.bell-labs.com/who/sjf/
# - Python translation to file voronoi.py by Bill Simons, 2005, http://www.oxfish.com/
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Create Voronoi diagram from seeds (midpoints of selected objects)
"""

import random

import inkex
from inkex import Group, Rectangle, PathElement, Vector2d as Point

import voronoi


class Voronoi(inkex.EffectExtension):
    """Extension to create a Voronoi diagram."""

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument(
            "--diagram-type",
            default="Voronoi",
            dest="diagramType",
            choices=["Voronoi", "Delaunay", "Both"],
            help="Defines the type of the diagram",
        )
        pars.add_argument(
            "--clip-box",
            choices=["Page", "Automatic from seeds"],
            default="Page",
            dest="clip_box",
            help="Defines the bounding box of the Voronoi diagram",
        )
        pars.add_argument(
            "--show-clip-box",
            type=inkex.Boolean,
            default=False,
            dest="showClipBox",
            help="Set this to true to write the bounding box",
        )
        pars.add_argument(
            "--delaunay-fill-options",
            default="delaunay-no-fill",
            dest="delaunayFillOptions",
            help="Set the Delaunay triangles color options",
        )

    def dot(self, x, y):
        """Clipping a line by a bounding box"""
        return x[0] * y[0] + x[1] * y[1]

    def intersect_line_segment(self, line, vt1, vt2):
        """Get the line intersection of the two verticies"""
        sc1 = self.dot(line, vt1) - line[2]
        sc2 = self.dot(line, vt2) - line[2]
        if sc1 * sc2 > 0:
            return 0, 0, False

        tmp = self.dot(line, vt1) - self.dot(line, vt2)
        if tmp == 0:
            return 0, 0, False
        und = (line[2] - self.dot(line, vt2)) / tmp
        vt0 = 1 - und
        return und * vt1[0] + vt0 * vt2[0], und * vt1[1] + vt0 * vt2[1], True

    def clip_edge(self, vertices, lines, edge, bbox):
        # bounding box corners
        bbc = [
            (bbox[0], bbox[2]),
            (bbox[1], bbox[2]),
            (bbox[1], bbox[3]),
            (bbox[0], bbox[3]),
        ]

        # record intersections of the line with bounding box edges
        if edge[0] >= len(lines):
            return []
        line = lines[edge[0]]
        interpoints = []
        for i in range(4):
            pnt = self.intersect_line_segment(line, bbc[i], bbc[(i + 1) % 4])
            if pnt[2]:
                interpoints.append(pnt)

        # if the edge has no intersection, return empty intersection
        if len(interpoints) < 2:
            return []

        if len(interpoints) > 2:  # happens when the edge crosses the corner of the box
            interpoints = list(set(interpoints))  # remove doubles

        # points of the edge
        vt1 = vertices[edge[1]]
        interpoints.append((vt1[0], vt1[1], False))
        vt2 = vertices[edge[2]]
        interpoints.append((vt2[0], vt2[1], False))

        # sorting the points in the widest range to get them in order on the line
        minx = interpoints[0][0]
        miny = interpoints[0][1]
        maxx = interpoints[0][0]
        maxy = interpoints[0][1]
        for point in interpoints:
            minx = min(point[0], minx)
            maxx = max(point[0], maxx)
            miny = min(point[1], miny)
            maxy = max(point[1], maxy)

        if (maxx - minx) > (maxy - miny):
            interpoints.sort()
        else:
            interpoints.sort(key=lambda pt: pt[1])

        start = []
        inside = False  # true when the part of the line studied is in the clip box
        start_write = False  # true when the part of the line is in the edge segment
        for point in interpoints:
            if point[2]:  # The point is a bounding box intersection
                if inside:
                    if start_write:
                        return [[start[0], start[1]], [point[0], point[1]]]
                    return []
                else:
                    if start_write:
                        start = point
                inside = not inside
            else:  # The point is a segment endpoint
                if start_write:
                    if inside:
                        # a vertex ends the line inside the bounding box
                        return [[start[0], start[1]], [point[0], point[1]]]
                    return []
                else:
                    if inside:
                        start = point
                start_write = not start_write

    def effect(self):
        # Check that elements have been selected
        if not self.svg.selection:
            inkex.errormsg(_("Please select objects!"))
            return

        linestyle = {
            "stroke": "#000000",
            "stroke-width": str(self.svg.to_dimensionless("1px")),
            "fill": "none",
            "stroke-linecap": "round",
            "stroke-linejoin": "round",
        }

        facestyle = {
            "stroke": "#000000",
            "stroke-width": str(self.svg.to_dimensionless("1px")),
            "fill": "none",
            "stroke-linecap": "round",
            "stroke-linejoin": "round",
        }

        parent_group = self.svg.selection.first().getparent()
        trans = parent_group.composed_transform()

        invtrans = None
        if trans:
            invtrans = -trans

        # Recovery of the selected objects
        pts = []
        nodes = []
        seeds = []
        fills = []

        for node in self.svg.selection.values():
            nodes.append(node)
            bbox = node.bounding_box()
            if bbox:
                center_x, center_y = bbox.center
                point = [center_x, center_y]
                if trans:
                    point = trans.apply_to_point(point)
                pts.append(Point(*point))
                if self.options.delaunayFillOptions != "delaunay-no-fill":
                    fills.append(node.style.get("fill", "none"))
                seeds.append(Point(center_x, center_y))

        # Creation of groups to store the result
        if self.options.diagramType != "Delaunay":
            # Voronoi
            group_voronoi = parent_group.add(Group())
            group_voronoi.set("inkscape:label", "Voronoi")
            if invtrans:
                group_voronoi.transform @= invtrans
        if self.options.diagramType != "Voronoi":
            # Delaunay
            group_delaunay = parent_group.add(Group())
            group_delaunay.set("inkscape:label", "Delaunay")

        # Clipping box handling
        if self.options.diagramType != "Delaunay":
            # Clipping bounding box creation
            group_bbox = sum([node.bounding_box() for node in nodes], None)

            # Clipbox is the box to which the Voronoi diagram is restricted
            if self.options.clip_box == "Page":
                width = self.svg.viewbox_width
                height = self.svg.viewbox_height
                clip_box = (0, width, 0, height)
            else:
                clip_box = (
                    group_bbox.left,
                    group_bbox.right,
                    group_bbox.top,
                    group_bbox.bottom,
                )

            # Safebox adds points so that no Voronoi edge in clip_box is infinite
            safe_box = (
                2 * clip_box[0] - clip_box[1],
                2 * clip_box[1] - clip_box[0],
                2 * clip_box[2] - clip_box[3],
                2 * clip_box[3] - clip_box[2],
            )
            pts.append(Point(safe_box[0], safe_box[2]))
            pts.append(Point(safe_box[1], safe_box[2]))
            pts.append(Point(safe_box[1], safe_box[3]))
            pts.append(Point(safe_box[0], safe_box[3]))

            if self.options.showClipBox:
                # Add the clip box to the drawing
                rect = group_voronoi.add(Rectangle())
                rect.set("x", str(clip_box[0]))
                rect.set("y", str(clip_box[2]))
                rect.set("width", str(clip_box[1] - clip_box[0]))
                rect.set("height", str(clip_box[3] - clip_box[2]))
                rect.style = linestyle

        # Voronoi diagram generation
        if self.options.diagramType != "Delaunay":
            vertices, lines, edges = voronoi.computeVoronoiDiagram(pts)
            for edge in edges:
                vindex1, vindex2 = edge[1:]
                if (vindex1 < 0) or (vindex2 < 0):
                    continue  # infinite lines have no need to be handled in the clipped box
                else:
                    segment = self.clip_edge(vertices, lines, edge, clip_box)
                    # segment = [vertices[vindex1],vertices[vindex2]] # deactivate clipping
                    if len(segment) > 1:
                        x1, y1 = segment[0]
                        x2, y2 = segment[1]
                        cmds = [["M", [x1, y1]], ["L", [x2, y2]]]
                        path = group_voronoi.add(PathElement())
                        path.set("d", str(inkex.Path(cmds)))
                        path.style = linestyle

        if self.options.diagramType != "Voronoi":
            triangles = voronoi.computeDelaunayTriangulation(seeds)
            i = 0
            if self.options.delaunayFillOptions == "delaunay-fill":
                random.seed("inkscape")
            for triangle in triangles:
                pt1 = seeds[triangle[0]]
                pt2 = seeds[triangle[1]]
                pt3 = seeds[triangle[2]]
                cmds = [
                    ["M", [pt1.x, pt1.y]],
                    ["L", [pt2.x, pt2.y]],
                    ["L", [pt3.x, pt3.y]],
                    ["Z", []],
                ]
                if (
                    self.options.delaunayFillOptions == "delaunay-fill"
                    or self.options.delaunayFillOptions == "delaunay-fill-random"
                ):
                    facestyle = {
                        "stroke": fills[triangle[random.randrange(0, 2)]],
                        "stroke-width": str(self.svg.to_dimensionless("0.005px")),
                        "fill": fills[triangle[random.randrange(0, 2)]],
                        "stroke-linecap": "round",
                        "stroke-linejoin": "round",
                    }
                path = group_delaunay.add(PathElement())
                path.set("d", str(inkex.Path(cmds)))
                path.style = facestyle
                i += 1


if __name__ == "__main__":
    Voronoi().run()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Thomas Holder <thomas.holder@schrodinger.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Interface for the Use and Symbol elements
"""

from ..transforms import Transform

from ._groups import Group, GroupBase
from ._base import ShapeElement


class Symbol(GroupBase):
    """SVG symbol element"""

    tag_name = "symbol"


class Use(ShapeElement):
    """A 'use' element that links to another in the document"""

    tag_name = "use"

    @classmethod
    def new(cls, elem, x, y, **attrs):  # pylint: disable=arguments-differ
        ret = super().new(x=x, y=y, **attrs)
        ret.href = elem
        return ret

    def get_path(self):
        """Returns the path of the cloned href plus any transformation"""
        path = self.href.path
        path.transform(self.href.transform)
        return path

    def effective_style(self):
        """Href's style plus this object's own styles"""
        style = self.href.effective_style()
        style.update(self.style)
        return style

    def unlink(self):
        """Unlink this clone, replacing it with a copy of the original"""
        copy = self.href.copy()
        if isinstance(copy, Symbol):
            group = Group(**copy.attrib)
            group.extend(copy)
            copy = group
        copy.transform = self.transform @ copy.transform
        copy.transform.add_translate(
            self.to_dimensionless(self.get("x", 0)),
            self.to_dimensionless(self.get("y", 0)),
        )
        copy.style = self.style + copy.style
        self.replace_with(copy)
        copy.set_random_ids()
        return copy

    def shape_box(self, transform=None):
        """BoundingBox of the unclipped shape

        .. versionadded:: 1.1"""
        effective_transform = Transform(transform) @ self.transform
        return self.href.bounding_box(effective_transform)



# coding=utf-8
# COPYRIGHT
#
# pylint: disable=invalid-name
#
"""
Depreicated simplepath replacements with documentation
"""

import math
from inkex.deprecated import deprecate, DeprecatedDict
from inkex.transforms import Transform
from inkex.paths import Path

pathdefs = DeprecatedDict(
    {
        "M": ["L", 2, [float, float], ["x", "y"]],
        "L": ["L", 2, [float, float], ["x", "y"]],
        "H": ["H", 1, [float], ["x"]],
        "V": ["V", 1, [float], ["y"]],
        "C": [
            "C",
            6,
            [float, float, float, float, float, float],
            ["x", "y", "x", "y", "x", "y"],
        ],
        "S": ["S", 4, [float, float, float, float], ["x", "y", "x", "y"]],
        "Q": ["Q", 4, [float, float, float, float], ["x", "y", "x", "y"]],
        "T": ["T", 2, [float, float], ["x", "y"]],
        "A": [
            "A",
            7,
            [float, float, float, int, int, float, float],
            ["r", "r", "a", 0, "s", "x", "y"],
        ],
        "Z": ["L", 0, [], []],
    }
)


@deprecate
def parsePath(d):
    """element.path.to_arrays()"""
    return Path(d).to_arrays()


@deprecate
def formatPath(a):
    """str(element.path) or str(Path(array))"""
    return str(Path(a))


@deprecate
def translatePath(p, x, y):
    """Path(array).translate(x, y)"""
    p[:] = Path(p).translate(x, y).to_arrays()


@deprecate
def scalePath(p, x, y):
    """Path(array).scale(x, y)"""
    p[:] = Path(p).scale(x, y).to_arrays()


@deprecate
def rotatePath(p, a, cx=0, cy=0):
    """Path(array).rotate(angle_degrees, (center_x, center_y))"""
    p[:] = Path(p).rotate(math.degrees(a), (cx, cy)).to_arrays()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2011 Felipe Correa da Silva Sanches
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

from typing import Union

import inkex
from inkex.localization import inkex_gettext as _


class SetupTypographyCanvas(inkex.EffectExtension):
    """Resizes the canvas and creates typography-relevant guidelines"""

    def add_arguments(self, pars):
        pars.add_argument("-e", "--emsize", type=int, default=1000)
        pars.add_argument("-c", "--caps", type=int, default=700, help="Caps Height")
        pars.add_argument("-x", "--xheight", type=int, default=500)
        pars.add_argument("-a", "--ascender", type=int, default=750)
        pars.add_argument("-d", "--descender", type=int, default=250)

    def create_horizontal_guideline(
        self, name: str, position: Union[int, float]
    ) -> inkex.BaseElement:
        """Create a horizontal guideline with name and position

        Args:
            name (str): the name of the guideline
            position (Union[int, float]): the vertical position of the guideline

        Returns:
            inkex.BaseElement: the created guideline
        """
        return self.svg.namedview.add(
            inkex.Guide().move_to(0, position, (0, 1)).update(inkscape__label=name)
        )

    def effect(self):
        # Get all the options
        emsize = self.options.emsize
        ascender = self.options.ascender
        caps = self.options.caps
        xheight = self.options.xheight
        descender = self.options.descender

        # Get access to main SVG document element
        self.svg.set("width", str(emsize))
        self.svg.set("height", str(emsize))
        self.svg.set("viewBox", "0 0 " + str(emsize) + " " + str(emsize))

        baseline = descender
        # Create guidelines
        self.create_horizontal_guideline(_("baseline"), baseline)
        self.create_horizontal_guideline(_("ascender"), baseline + ascender)
        self.create_horizontal_guideline(_("caps"), baseline + caps)
        self.create_horizontal_guideline(_("xheight"), baseline + xheight)
        self.create_horizontal_guideline(_("descender"), baseline - descender)

        namedview = self.svg.namedview
        namedview.set("inkscape:document-units", "px")
        namedview.set("inkscape:cx", str(emsize / 2.0))
        namedview.set("inkscape:cy", str(emsize / 2.0))


if __name__ == "__main__":
    SetupTypographyCanvas().run()



# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#               2019-2020 Martin Owens
#               2021 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Functions for handling styles and embedded css
"""

import re
from collections import OrderedDict
from typing import MutableMapping, Union, Iterable, TYPE_CHECKING

from .interfaces.IElement import IBaseElement

from .colors import Color
from .properties import BaseStyleValue, all_properties, ShorthandValue
from .css import ConditionalRule

if TYPE_CHECKING:
    from .elements._svg import SvgDocumentElement


class Classes(list):
    """A list of classes applied to an element (used in css and js)"""

    def __init__(self, classes=None, callback=None):
        self.callback = None
        if isinstance(classes, str):
            classes = classes.split()
        super().__init__(classes or ())
        self.callback = callback

    def __str__(self):
        return " ".join(self)

    def _callback(self):
        if self.callback is not None:
            self.callback(self)

    def __setitem__(self, index, value):
        super().__setitem__(index, value)
        self._callback()

    def append(self, value):
        value = str(value)
        if value not in self:
            super().append(value)
            self._callback()

    def remove(self, value):
        value = str(value)
        if value in self:
            super().remove(value)
            self._callback()

    def toggle(self, value):
        """If exists, remove it, if not, add it"""
        value = str(value)
        if value in self:
            return self.remove(value)
        return self.append(value)


class Style(OrderedDict, MutableMapping[str, Union[str, BaseStyleValue]]):
    """A list of style directives

    .. versionchanged:: 1.2
        The Style API now allows for access to parsed / processed styles via the
        :func:`call` method.

    .. automethod:: __call__
    .. automethod:: __getitem__
    .. automethod:: __setitem__
    """

    color_props = ("stroke", "fill", "stop-color", "flood-color", "lighting-color")
    opacity_props = ("stroke-opacity", "fill-opacity", "opacity", "stop-opacity")
    unit_props = "stroke-width"
    """Dictionary of attributes with units. 
    
    ..versionadded:: 1.2
    """
    associated_props = {
        "fill": "fill-opacity",
        "stroke": "stroke-opacity",
        "stop-color": "stop-opacity",
    }
    """Dictionary of association between color and opacity attributes.

    .. versionadded:: 1.2
    """

    def __init__(self, style=None, callback=None, element=None, **kw):
        self.element = element
        # This callback is set twice because this is 'pre-initial' data (no callback)
        self.callback = None
        # Either a string style or kwargs (with dashes as underscores).
        style = style or [(k.replace("_", "-"), v) for k, v in kw.items()]
        if isinstance(style, str):
            style = self._parse_str(style)
        # Order raw dictionaries so tests can be made reliable
        if isinstance(style, dict) and not isinstance(style, OrderedDict):
            style = [(name, style[name]) for name in sorted(style)]
        # Should accept dict, Style, parsed string, list etc.
        super().__init__(style)
        # Now after the initial data, the callback makes sense.
        self.callback = callback

    @staticmethod
    def _parse_str(style: str, element=None) -> Iterable[BaseStyleValue]:
        """Create a dictionary from the value of a CSS rule (such as an inline style or
        from an embedded style sheet), including its !important state, parsing the value
        if possible.

        Args:
            style: the content of a CSS rule to parse
            element: the element this style is working on (can be the root SVG, is used
                for parsing gradients etc.)

        Yields:
            :class:`~inkex.properties.BaseStyleValue`: the parsed attribute
        """
        for declaration in style.split(";"):
            if ":" in declaration:
                result = BaseStyleValue.factory_errorhandled(
                    element, declaration=declaration.strip()
                )
                if result is not None:
                    yield result

    @staticmethod
    def parse_str(style: str, element=None):
        """Parse a style passed as string"""
        return Style(style, element=element)

    def __str__(self):
        """Format an inline style attribute from a dictionary"""
        return self.to_str()

    def to_str(self, sep=";"):
        """Convert to string using a custom delimiter"""
        return sep.join([self.get_store(key).declaration for key in self])

    def __add__(self, other):
        """Add two styles together to get a third, composing them"""
        ret = self.copy()
        ret.update(Style(other))
        return ret

    def __iadd__(self, other):
        """Add style to this style, the same as ``style.update(dict)``"""
        self.update(other)
        return self

    def __sub__(self, other):
        """Remove keys and return copy"""
        ret = self.copy()
        ret.__isub__(other)
        return ret

    def __isub__(self, other):
        """Remove keys from this style, list of keys or other style dictionary"""
        for key in other:
            self.pop(key, None)
        return self

    def __ne__(self, other):
        return not self.__eq__(other)

    def copy(self):
        """Create a copy of the style.

        .. versionadded:: 1.2"""
        ret = Style({}, element=self.element)
        for key, value in super().items():
            ret[key] = value
        return ret

    def update(self, other):
        """Update, while respecting ``!important`` declarations."""
        if not isinstance(other, Style):
            other = Style(other)
        # only update
        if isinstance(other, Style):
            for key in other.keys():
                if not (self.get_importance(key) and not other.get_importance(key)):
                    self[key] = other.get_store(key)

        if self.callback is not None:
            self.callback(self)

    def add_inherited(self, parent):
        """Creates a new Style containing all parent styles with importance "!important"
        and current styles with importance "!important"

        .. versionadded:: 1.2

        Args:
            parent: the parent style that will be merged into this one (will not be
                altered)

        Returns:
            Style: the merged Style object
        """
        ret = self.copy()
        ret.apply_shorthands()  # parent should already have its shortcuts applied

        if not (isinstance(parent, Style)):
            return ret

        for key in parent.keys():
            apply = False
            if key in all_properties and all_properties[key][3]:
                # only set parent value if value is not set or parent importance is
                # higher
                if key not in ret:
                    apply = True
                elif self.get_importance(key) != parent.get_importance(key):
                    apply = parent.get_importance(key)
            if key in ret and ret[key] == "inherit":
                apply = True
            if apply:
                ret[key] = parent[key]
        return ret

    def apply_shorthands(self):
        """Apply all shorthands in this style."""
        for element in list(self.values()):
            if isinstance(element, ShorthandValue):
                element.apply_shorthand(self)

    def __delitem__(self, key):
        super().__delitem__(key)
        if self.callback is not None:
            self.callback(self)

    def __setitem__(self, key, value):
        """Sets a style value.

        .. versionchanged:: 1.2
            ``value`` can now also be non-string objects such as a Gradient.

        Args:
            key (str): the attribute name
            value (Any):

                - a :class:`BaseStyleValue`
                - a string with the value
                - any other object. The :class:`~inkex.properties.BaseStyleValue`
                  subclass of the provided key will attempt to create a string out of
                  the passed value.
        Raises:
            ValueError: when ``value`` is a :class:`~inkex.properties.BaseStyleValue`
                for a different attribute than `key`
            Error: Other exceptions may be raised when converting non-string objects."""
        if not isinstance(value, BaseStyleValue) or value is None:
            # try to convert the value using the factory
            value = BaseStyleValue.factory(attr_name=key, value=value)
            # check if the set attribute is valid
            _ = value.parse_value(self.element)
        elif key != value.attr_name:
            raise ValueError(
                """You're trying to save a value into a style attribute, but the
                provided key is different from the attribute name given in the value"""
            )
        super().__setitem__(key, value)
        if self.callback is not None:
            self.callback(self)

    def __getitem__(self, key):
        """Returns the unparsed value of the element (minus a possible ``!important``)

        .. versionchanged:: 1.2
            ``!important`` is removed from the value.
        """
        return self.get_store(key).value

    def get(self, key, default=None):
        if key in self:
            return self.__getitem__(key)
        return default

    def get_store(self, key):
        """Gets the :class:`~inkex.properties.BaseStyleValue` of this key, since the
        other interfaces - :func:`__getitem__` and :func:`__call__` - return the
        original and parsed value, respectively.

        .. versionadded:: 1.2

        Args:
            key (str): the attribute name

        Returns:
            BaseStyleValue: the BaseStyleValue struct of this attribute
        """
        return super().__getitem__(key)

    def __call__(self, key, element=None):
        """Return the parsed value of a style. Optionally, an element can be passed
        that will be used to find gradient definitions ect.

        .. versionadded:: 1.2"""
        # check if there are shorthand properties defined. If so, apply them to a copy
        copy = self
        for value in super().values():
            if isinstance(value, ShorthandValue):
                copy = self.copy()
                copy.apply_shorthands()
        if key in copy:
            return copy.get_store(key).parse_value(element or self.element)
        # style is not set, return the default value
        if key in all_properties:
            defvalue = BaseStyleValue.factory(
                attr_name=key, value=all_properties[key][1]
            )
            return (
                defvalue.parse_value()
            )  # default values are independent of the element
        raise KeyError("Unknown attribute")

    def __eq__(self, other):
        if not isinstance(other, Style):
            other = Style(other)
        if self.keys() != other.keys():
            return False
        for arg in set(self) | set(other):
            if self.get_store(arg) != other.get_store(arg):
                return False
        return True

    def items(self):
        """The styles's parsed items

        .. versionadded:: 1.2"""
        for key, value in super().items():
            yield key, value.value

    def get_importance(self, key, default=False):
        """Returns whether the declaration with ``key`` is marked as ``!important``

        .. versionadded:: 1.2"""
        if key in self:
            return super().__getitem__(key).important
        return default

    def set_importance(self, key, importance):
        """Sets the ``!important`` state of a declaration with key ``key``

        .. versionadded:: 1.2"""
        if key in self:
            super().__getitem__(key).important = importance
        else:
            raise KeyError()
        if self.callback is not None:
            self.callback(self)

    def get_color(self, name="fill"):
        """Get the color AND opacity as one Color object"""
        color = Color(self.get(name, "none"))
        return color.to_rgba(self.get(name + "-opacity", 1.0))

    def set_color(self, color, name="fill"):
        """Sets the given color AND opacity as rgba to the fill or stroke style
        properties."""
        color = Color(color)
        if color.space == "rgba" and name in Style.associated_props:
            self[Style.associated_props[name]] = color.alpha
            self[name] = color.to_rgb()
        else:
            self[name] = color

    def update_urls(self, old_id, new_id):
        """Find urls in this style and replace them with the new id"""
        for (name, value) in self.items():
            if value == f"url(#{old_id})":
                self[name] = f"url(#{new_id})"

    def interpolate(self, other, fraction):
        # type: (Style, Style, float) -> Style
        """Interpolate all properties.

        .. versionadded:: 1.1"""
        from .tween import StyleInterpolator
        from inkex.elements import PathElement

        if self.element is None:
            self.element = PathElement(style=str(self))
        if other.element is None:
            other.element = PathElement(style=str(other))
        return StyleInterpolator(self.element, other.element).interpolate(fraction)

    @classmethod
    def cascaded_style(cls, element):
        """Returns the cascaded style of an element (all rules that apply the element
        itself), based on the stylesheets, the presentation attributes and the inline
        style using the respective specificity of the style

        see https://www.w3.org/TR/CSS22/cascade.html#cascading-order

        .. versionadded:: 1.2

        Args:
            element (BaseElement): the element that the cascaded style will be
                computed for

        Returns:
            Style: the cascaded style
        """
        styles = list(element.root.stylesheets.lookup_specificity(element.get_id()))

        # presentation attributes have specificity 0,
        # see https://www.w3.org/TR/SVG/styling.html#PresentationAttributes
        styles.append([element.presentation_style(), (0, 0, 0)])

        # would be (1, 0, 0, 0), but then we'd have to extend every entry
        styles.append([element.style, (float("inf"), 0, 0)])

        # sort styles by specificity (ascending, so when overwriting it's correct)
        styles = sorted(styles, key=lambda item: item[1])

        result = styles[0][0].copy()
        for style, _ in styles[1:]:
            result.update(style)
        result.element = element
        return result

    @classmethod
    def specified_style(cls, element):
        """Returns the specified style of an element, i.e. the cascaded style +
        inheritance, see https://www.w3.org/TR/CSS22/cascade.html#specified-value

        .. versionadded:: 1.2

        Args:
            element (BaseElement): the element that the specified style will be computed
                for

        Returns:
            Style: the specified style
        """

        # We currently dont treat the case where parent=absolute value and
        # element=relative value, i.e. specified = relative * absolute.
        cascaded = Style.cascaded_style(element)

        parent = element.getparent()

        if parent is not None and isinstance(parent, IBaseElement):
            cascaded = Style.add_inherited(cascaded, parent.specified_style())
        cascaded.element = element
        return cascaded  # doesn't have a parent


class StyleSheets(list):
    """
    Special mechanism which contains all the stylesheets for an svg document
    while also caching lookups for specific elements.

    This caching is needed because data can't be attached to elements as they are
    re-created on the fly by lxml so lookups have to be centralised.
    """

    def __init__(self, svg=None):
        super().__init__()
        self.svg = svg

    def lookup(self, element_id, svg=None):
        """
        Find all styles for this element.
        """
        # This is aweful, but required because we can't know for sure
        # what might have changed in the xml tree.
        if svg is None:
            svg = self.svg
        for sheet in self:
            for style in sheet.lookup(element_id, svg=svg):
                yield style

    def lookup_specificity(self, element_id, svg=None):
        """
        Find all styles for this element and return the specificity of the match.

        .. versionadded:: 1.2
        """
        # This is aweful, but required because we can't know for sure
        # what might have changed in the xml tree.
        if svg is None:
            svg = self.svg
        for sheet in self:
            for style in sheet.lookup_specificity(element_id, svg=svg):
                yield style


class StyleSheet(list):
    """
    A style sheet, usually the CDATA contents of a style tag, but also
    a css file used with a css. Will yield multiple Style() classes.
    """

    comment_strip = re.compile(r"(\/\/.*?\n)|(\/\*.*?\*\/)|@.*;")

    def __init__(self, content=None, callback=None):
        super().__init__()
        self.callback = None
        # Remove comments
        content = self.comment_strip.sub("", (content or ""))
        # Parse rules
        for block in content.split("}"):
            if block:
                self.append(block)
        self.callback = callback

    def __str__(self):
        return "\n" + "\n".join([str(style) for style in self]) + "\n"

    def _callback(self, style=None):  # pylint: disable=unused-argument
        if self.callback is not None:
            self.callback(self)

    def add(self, rule, style):
        """Append a rule and style combo to this stylesheet"""
        self.append(
            ConditionalStyle(rules=rule, style=str(style), callback=self._callback)
        )

    def append(self, other):
        """Make sure callback is called when updating"""
        if isinstance(other, str):
            if "{" not in other:
                return  # Warning?
            rules, style = other.strip("}").split("{", 1)
            if rules.strip().startswith("@"):  # ignore @font-face and @import
                return
            other = ConditionalStyle(
                rules=rules, style=style.strip(), callback=self._callback
            )
        super().append(other)
        self._callback()

    def lookup(self, element_id, svg):
        """Lookup the element_id against all the styles in this sheet"""
        for style in self:
            for elem in svg.xpath(style.to_xpath()):
                if elem.get("id", None) == element_id:
                    yield style

    def lookup_specificity(self, element_id, svg):
        """Lookup the element_id against all the styles in this sheet
        and return the specificity of the match

        Args:
            element_id (str): the id of the element that styles are being queried for
            svg (SvgDocumentElement): The document that contains both element and the
                styles

        Yields:
            Tuple[ConditionalStyle, Tuple[int, int, int]]: all matched styles and the
            specificity of the match
        """
        for style in self:
            for rule, spec in zip(style.to_xpaths(), style.get_specificities()):
                for elem in svg.xpath(rule):
                    if elem.get("id", None) == element_id:
                        yield (style, spec)


class ConditionalStyle(Style):
    """
    Just like a Style object, but includes one or more
    conditional rules which places this style in a stylesheet
    rather than being an attribute style.
    """

    def __init__(self, rules="*", style=None, callback=None, **kwargs):
        super().__init__(style=style, callback=callback, **kwargs)
        self.rules = [ConditionalRule(rule) for rule in rules.split(",")]

    def __str__(self):
        """Return this style as a css entry with class"""
        content = self.to_str(";\n  ")
        rules = ",\n".join(str(rule) for rule in self.rules)
        if content:
            return f"{rules} {{\n  {content};\n}}"
        return f"{rules} {{}}"

    def to_xpath(self):
        """Convert all rules to an xpath"""
        # This can be converted to cssselect.CSSSelector (lxml.cssselect) later if we
        # have coverage problems. The main reason we're not is that cssselect is doing
        # exactly this xpath transform and provides no extra functionality for reverse
        # lookups.
        return "|".join(self.to_xpaths())

    def to_xpaths(self):
        """Gets a list of xpaths for all rules of this ConditionalStyle

        .. versionadded:: 1.2"""
        return [rule.to_xpath() for rule in self.rules]

    def get_specificities(self):
        """Gets an iterator of the specificity of all rules in this ConditionalStyle

        .. versionadded:: 1.2"""
        for rule in self.rules:
            yield rule.get_specificity()



# coding=utf-8
#
# Copyright (C) 2019 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
This API provides methods for calling Inkscape to execute a given
Inkscape command. This may be needed for various compiling options
(e.g., png), running other extensions or performing other options only
available via the shell API.

Best practice is to avoid using this API except when absolutely necessary,
since it is resource-intensive to invoke a new Inkscape instance.

However, in any circumstance when it is necessary to call Inkscape, it
is strongly recommended that you do so through this API, rather than calling
it yourself, to take advantage of the security settings and testing functions.

"""

import os
import sys
from shutil import which as warlock

from subprocess import Popen, PIPE
from tempfile import TemporaryDirectory
from typing import List
from lxml.etree import ElementTree

from .elements import SvgDocumentElement

INKSCAPE_EXECUTABLE_NAME = os.environ.get("INKSCAPE_COMMAND")
if INKSCAPE_EXECUTABLE_NAME is None:
    if sys.platform == "win32":
        # prefer inkscape.exe over inkscape.com which spawns a command window
        INKSCAPE_EXECUTABLE_NAME = "inkscape.exe"
    else:
        INKSCAPE_EXECUTABLE_NAME = "inkscape"


class CommandNotFound(IOError):
    """Command is not found"""


class ProgramRunError(ValueError):
    """A specialized ValueError that is raised when a call to an external command fails.
    It stores additional information about a failed call to an external program.

    If only the ``program`` parameter is given, it is interpreted as the error message.
    Otherwise, the error message is compiled from all constructor parameters."""

    program: str
    """The absolute path to the called executable"""

    returncode: int
    """Return code of the program call"""

    stderr: str
    """stderr stream output of the call"""

    stdout: str
    """stdout stream output of the call"""

    arguments: List
    """Arguments of the call"""

    def __init__(self, program, returncode=None, stderr=None, stdout=None, args=None):
        self.program = program
        self.returncode = returncode
        self.stderr = stderr
        self.stdout = stdout
        self.arguments = args
        super().__init__(str(self))

    def __str__(self):
        if self.returncode is None:
            return self.program
        return (
            f"Return Code: {self.returncode}: {self.stderr}\n{self.stdout}"
            f"\nargs: {self.args}"
        )


def which(program):
    """
    Attempt different methods of trying to find if the program exists.
    """
    if os.path.isabs(program) and os.path.isfile(program):
        return program
    # On Windows, shutil.which may give preference to .py files in the current directory
    # (such as pdflatex.py), e.g. if .PY is in pathext, because the current directory is
    # prepended to PATH. This can be suppressed by explicitly appending the current
    # directory.

    try:
        if sys.platform == "win32":
            prog = warlock(program, path=os.environ["PATH"] + ";" + os.curdir)
            if prog:
                return prog
    except ImportError:
        pass

    try:
        # Python3 only version of which
        prog = warlock(program)
        if prog:
            return prog
    except ImportError:
        pass  # python2

    # There may be other methods for doing a `which` command for other
    # operating systems; These should go here as they are discovered.

    raise CommandNotFound(f"Can not find the command: '{program}'")


def write_svg(svg, *filename):
    """Writes an svg to the given filename"""
    filename = os.path.join(*filename)
    if os.path.isfile(filename):
        return filename
    with open(filename, "wb") as fhl:
        if isinstance(svg, SvgDocumentElement):
            svg = ElementTree(svg)
        if hasattr(svg, "write"):
            # XML document
            svg.write(fhl)
        elif isinstance(svg, bytes):
            fhl.write(svg)
        else:
            raise ValueError("Not sure what type of SVG data this is.")
    return filename


def to_arg(arg, oldie=False):
    """Convert a python argument to a command line argument"""
    if isinstance(arg, (tuple, list)):
        (arg, val) = arg
        arg = "-" + arg
        if len(arg) > 2 and not oldie:
            arg = "-" + arg
        if val is True:
            return arg
        if val is False:
            return None
        return f"{arg}={str(val)}"
    return str(arg)


def to_args(prog, *positionals, **arguments):
    """Compile arguments and keyword arguments into a list of strings which Popen will
    understand.

    :param prog:
        Program executable prepended to the output.
    :type first: ``str``

    :Arguments:
        * (``str``) -- String added as given
        * (``tuple``) -- Ordered version of Kwyward Arguments, see below

    :Keyword Arguments:
        * *name* (``str``) --
          Becomes ``--name="val"``
        * *name* (``bool``) --
          Becomes ``--name``
        * *name* (``list``) --
          Becomes ``--name="val1"`` ...
        * *n* (``str``) --
          Becomes ``-n=val``
        * *n* (``bool``) --
          Becomes ``-n``

    :return: Returns a list of compiled arguments ready for Popen.
    :rtype: ``list[str]``
    """
    args = [prog]
    oldie = arguments.pop("oldie", False)
    for arg, value in arguments.items():
        arg = arg.replace("_", "-").strip()

        if isinstance(value, tuple):
            value = list(value)
        elif not isinstance(value, list):
            value = [value]

        for val in value:
            args.append(to_arg((arg, val), oldie))

    args += [to_arg(pos, oldie) for pos in positionals if pos is not None]
    # Filter out empty non-arguments
    return [arg for arg in args if arg is not None]


def to_args_sorted(prog, *positionals, **arguments):
    """same as :func:`to_args`, but keyword arguments are sorted beforehand

    .. versionadded:: 1.2"""
    return to_args(prog, *positionals, **dict(sorted(arguments.items())))


def _call(program, *args, **kwargs):
    stdin = kwargs.pop("stdin", None)
    if isinstance(stdin, str):
        stdin = stdin.encode("utf-8")
    inpipe = PIPE if stdin else None

    args = to_args(which(program), *args, **kwargs)

    kwargs = {}
    if sys.platform == "win32":
        kwargs["creationflags"] = 0x08000000  # create no console window

    with Popen(
        args,
        shell=False,  # Never have shell=True
        stdin=inpipe,  # StdIn not used (yet)
        stdout=PIPE,  # Grab any output (return it)
        stderr=PIPE,  # Take all errors, just incase
        **kwargs,
    ) as process:
        (stdout, stderr) = process.communicate(input=stdin)
        if process.returncode == 0:
            return stdout
        raise ProgramRunError(program, process.returncode, stderr, stdout, args)


def call(program, *args, **kwargs):
    """
    Generic caller to open any program and return its stdout::

        stdout = call('executable', arg1, arg2, dash_dash_arg='foo', d=True, ...)

    Will raise :class:`ProgramRunError` if return code is not 0.

    Keyword arguments:
        return_binary: Should stdout return raw bytes (default: False)

            .. versionadded:: 1.1
        stdin: The string or bytes containing the stdin (default: None)

    All other arguments converted using :func:`to_args` function.
    """
    # We use this long input because it's less likely to conflict with --binary=
    binary = kwargs.pop("return_binary", False)
    stdout = _call(program, *args, **kwargs)
    # Convert binary to string when we wish to have strings we do this here
    # so the mock tests will also run the conversion (always returns bytes)
    if not binary and isinstance(stdout, bytes):
        return stdout.decode(sys.stdout.encoding or "utf-8")
    return stdout


def inkscape(svg_file, *args, **kwargs):
    """
    Call Inkscape with the given svg_file and the given arguments, see call()
    """
    return call(INKSCAPE_EXECUTABLE_NAME, svg_file, *args, **kwargs)


def inkscape_command(svg, select=None, verbs=()):
    """
    Executes a list of commands, a mixture of verbs, selects etc.

    inkscape_command('<svg...>', ('verb', 'VerbName'), ...)
    """
    with TemporaryDirectory(prefix="inkscape-command") as tmpdir:
        svg_file = write_svg(svg, tmpdir, "input.svg")
        select = ("select", select) if select else None
        verbs += ("FileSave", "FileQuit")
        inkscape(svg_file, select, batch_process=True, verb=";".join(verbs))
        with open(svg_file, "rb") as fhl:
            return fhl.read()


def take_snapshot(svg, dirname, name="snapshot", ext="png", dpi=96, **kwargs):
    """
    Take a snapshot of the given svg file.

    Resulting filename is yielded back, after generator finishes, the
    file is deleted so you must deal with the file inside the for loop.
    """
    svg_file = write_svg(svg, dirname, name + ".svg")
    ext_file = os.path.join(dirname, name + "." + str(ext).lower())
    inkscape(
        svg_file, export_dpi=dpi, export_filename=ext_file, export_type=ext, **kwargs
    )
    return ext_file


def is_inkscape_available():
    """Return true if the Inkscape executable is available."""
    try:
        return bool(which(INKSCAPE_EXECUTABLE_NAME))
    except CommandNotFound:
        return False
